from functools import partial

# from typing import Callable
from collections.abc import Callable
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import LinearRegression
from sklearn.cluster import KMeans
from front import Crudifier

mall = dict(
    func=dict(StandardScaler=StandardScaler, KMeans=KMeans, linreg=LinearRegression),
    parametrized_funcs=dict(),
)

crudify = partial(Crudifier, mall=mall)


def str_to_dict(string):
    import json

    return json.loads(string)


def dict_maker(**kwargs):
    return dict(**kwargs)


@crudify('func', output_store='parametrized_funcs')
def factory(func, kwargs='()'):
    kwargs = str_to_dict(kwargs)
    return partial(func, **kwargs)


def foo(x):
    return x + 1


if __name__ == '__main__':
    from streamlitfront import mk_app
    from front.util import deep_merge
    from streamlitfront import dispatch_funcs

    # app = dispatch_funcs([factory])

    import front as f
    import streamlitfront.elements as sf

    # my_convention = {
    #     # f.APP_KEY: {'title': 'apply_func_to_wf'},
    #     f.RENDERING_KEY: {
    #         Callable: {
    #             'execution': {
    #                 'inputs': {
    #                     'func': {
    #                         f.ELEMENT_KEY: sf.SelectBox,
    #                         'options': list(mall['func'])
    #                     },
    #                 },
    #             }
    #         },
    #     }
    # }

    my_config = {
        # f.APP_KEY: {'title': 'apply_func_to_wf'},
        f.RENDERING_KEY: {
            'factory': {
                'execution': {
                    'inputs': {
                        'func': {
                            f.ELEMENT_KEY: sf.SelectBox,
                            'options': list(mall['func']),
                        },
                    },
                }
            },
        }
    }

    from i2 import Sig

    @Sig(KMeans)
    def kmeans(*args, **kwargs):
        return KMeans(*args, **kwargs)

    def change_types_to_callable(x):
        if isinstance(x, type):
            x = partial(x)
        return x

    app = mk_app([factory, kmeans], config=my_config,)
    app()

    import streamlit as st

    st.write(mall)



# launch as : opyrator launch-ui opyrator_front_dispatch:wrapped_func
# launch as : opyrator launch-api opyrator_front_dispatch:wrapped_func


import streamlit as st
import streamlit_pydantic as sp
from streamlitfront.base import dispatch_funcs, BasePageFunc
from opyratorfront.py2pydantic import func_to_pyd_input_model_cls
from pydantic import BaseModel, Field
from i2 import name_of_obj
from opyrator import Opyrator


# class Input(BaseModel):
#    inp: int


class Output(BaseModel):
    res: str


def func(x: str) -> str:
    result = x * 2
    return result


Input = func_to_pyd_input_model_cls(func, name='Input')
print('issubclass', issubclass(Input, BaseModel))
name = name_of_obj(func)


def wrapped_func(input: Input) -> Output:
    return Output(res=func(input.x))


# op = Opyrator(wrapped_func)



from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY
from collections.abc import Callable

from streamlitfront import mk_app, binder as b
from streamlitfront.elements import SelectBox
from typing import Any
from streamlitfront.elements import MultiSourceInput
import streamlit as st
from front.elements import InputBase
from i2 import Sig
from dataclasses import dataclass
from streamlitfront.elements import TextInput, SelectBox, ExecSection, TextOutput
from functools import partial
from front.crude import Crudifier
from streamlitfront.data_binding import BoundData


# # ============ MALL ===============
# if not b.mall():
#     b.mall = dict(func_store={})

# mall = b.mall()


def foo(x: str, y: str, z: int):
    return x + y * z


def bar(msg: str):
    return msg


# Choice #1: Simple ####################
# data = ['foo', 'bar']
# metadata = {'foo': foo, 'bar': bar}

# Choice #2: Involving a local file store ####################
from dol import Files

data = Files('~', max_levels=0)
metadata = {k: [foo, bar][len(k) % 2] for k in data}

# print(metadata)

# @Crudifier(output_store="func_store", mall=mall)
def my_map(func, kwargs):
    f = metadata[func]
    return partial(f, **kwargs)
    # return f, kwargs


def expand_names(names):
    return {name: {ELEMENT_KEY: TextInput} for name in names}


def populate_kwargs():
    value = b.selected_func()
    list_names = list(Sig(metadata[value]).names)
    b.names_for_kwargs = expand_names(list_names)
    st.write(b.names_for_kwargs())


if not b.selected_func():
    b.selected_func = next(iter(data))


def get_kwargs(**kwargs):
    return kwargs


@dataclass
class KwargsInput(InputBase):
    func_sig: Sig = None

    def __post_init__(self):
        super().__post_init__()
        self.get_kwargs = self.func_sig(get_kwargs)
        self.inputs = self._build_inputs_from_sig()

    def render(self):
        exec_section = ExecSection(
            obj=self.get_kwargs,
            inputs=self.inputs,
            output={ELEMENT_KEY: TextOutput},
            auto_submit=True,
            on_submit=self._return_kwargs,
            use_expander=False,
        )
        exec_section()
        return self.value()

    def _build_inputs_from_sig(self):
        return {
            name: {ELEMENT_KEY: TextInput, 'bound_data_factory': BoundData}
            for name in self.func_sig
        }

    def _return_kwargs(self, output):
        self.value.set(output)


if __name__ == '__main__':
    app = mk_app(
        [my_map],
        config={
            APP_KEY: {'title': 'Rendering map'},
            RENDERING_KEY: {
                'my_map': {
                    'execution': {
                        'inputs': {
                            'func': {
                                ELEMENT_KEY: SelectBox,
                                'options': data,
                                'value': b.selected_func,
                                'on_value_change': populate_kwargs,
                            },
                            'kwargs': {
                                ELEMENT_KEY: KwargsInput,
                                'func_sig': Sig(metadata[b.selected_func()]),
                            },
                        }
                    }
                },
            },
        },
    )
    app()



from front.spec_maker_base import APP_KEY, RENDERING_KEY, ELEMENT_KEY
from streamlitfront.elements import FloatSliderInput, TextOutput
from front.elements import OutputBase
from streamlitfront.base import mk_app
import streamlit as st
from st_aggrid import GridOptionsBuilder, AgGrid, DataReturnMode, GridUpdateMode
import pandas as pd
from front.crude import Crudifier
from functools import partial


d = {'A': [1, 2, 3], 'B': [4, 4, 3]}
df = pd.DataFrame(d)


class SimpleDfOutput(OutputBase):
    def render(self):
        st.write(self.output)


class TableOutput(OutputBase):
    def render(self):
        table_show(self.output)
        # st.write(self.output)


def display_df(df):
    return df


def configure_grid(gb):
    selection_mode = 'multiple'
    groupSelectsChildren = True
    groupSelectsFiltered = True
    gb.configure_selection(selection_mode)
    gb.configure_selection(
        selection_mode,
        use_checkbox=True,
        groupSelectsChildren=groupSelectsChildren,
        groupSelectsFiltered=groupSelectsFiltered,
    )
    return gb


def table_show(df):
    gb = GridOptionsBuilder.from_dataframe(df)

    gb = configure_grid(gb)
    gb.configure_grid_options(domLayout='normal')
    gridOptions = gb.build()

    return_mode = 'AS_INPUT'
    return_mode_value = DataReturnMode.__members__[return_mode]
    update_mode = 'MODEL_CHANGED'
    update_mode_value = GridUpdateMode.__members__[update_mode]

    grid_response = AgGrid(
        df,
        gridOptions=gridOptions,
        width='100%',
        data_return_mode=return_mode_value,
        update_mode=update_mode_value,
    )

    df_r = grid_response['data']
    st.write(df_r)
    selected = grid_response['selected_rows']
    df_selected = pd.DataFrame(selected).apply(pd.to_numeric, errors='coerce')
    st.write(df_selected)


mall = {'df_store': {'stored_wf': df}}


crudify = Crudifier(param_to_mall_map={'df': 'df_store'}, mall=mall)
f = crudify(display_df)


app = mk_app(
    [f],
    config={
        APP_KEY: {'title': 'My app'},
        RENDERING_KEY: {
            'f': {
                'execution': {
                    'output': {ELEMENT_KEY: TableOutput},
                    # "options": mall["df_store"],
                }
            },
        },
    },
)
app()



"""A simple app to demo the use of Mappings to handle complex type"""
from typing import Mapping, Callable, Any
from enum import Enum
from streamlitfront.page_funcs import (
    BasePageFunc,
    func_to_pyd_input_model_cls,
    name_of_obj,
)
import streamlit_pydantic as sp
import streamlit as st

# from front.py2pydantic import func_to_pyd_model_specs
from i2 import Sig, empty_param_attr
from pydantic import create_model, Field

ComplexType = float  # just pretend it's complex!


def func(salary: ComplexType, n_months: int = 12) -> float:
    return salary * n_months


SalaryKey = str  # or some type that will resolve in store-fed key selector
SalaryMapping = Mapping[SalaryKey, ComplexType]

salary_store: SalaryMapping
salary_store = {'sylvain': 10000, 'christian': 2000, 'thor': 50000}


def mk_choices_from_store(store):
    choices = Enum('Choices', {key: key for key in store.keys()})
    return choices


class SimplePageFuncPydanticWrite(BasePageFunc):
    def __call__(self, state):
        self.prepare_view(state)
        mymodel = func_to_pyd_input_model_cls(self.func)
        name = name_of_obj(self.func)
        data = sp.pydantic_form(key=f'my_form_{name}', model=mymodel)
        # data = sp.pydantic_input(key=f"my_form_{name}", model=mymodel)

        if data:
            st.write(self.func(**dict(data)))


def func_to_pyd_input_model_cls(func: Callable, dflt_type=Any, name=None):
    """Get a pydantic model of the arguments of a python function"""
    name = name or name_of_obj(func)
    return create_model(name, **dict(func_to_pyd_model_specs(func, dflt_type)))


def func_to_pyd_model_specs_with_tooltips(func: Callable, dflt_type=Any):
    """Helper function to get field info from python signature parameters"""
    text_info = func.__doc__
    for p in Sig(func).params:
        if p.annotation is not empty_param_attr:
            if p.default is not empty_param_attr:
                p.default = Field(p.default, description=text_info)
                yield p.name, (p.annotation, p.default)
            else:
                p.default = Field(..., description=text_info)
                yield p.name, (p.annotation, p.default)
        else:  # no annotations
            if p.default is not empty_param_attr:
                yield p.name, Field(p.default, description=text_info)
            else:
                yield p.name, Field(dflt_type, ..., description=text_info)


def func_to_pyd_model_specs(func: Callable, dflt_type=Any):
    """Helper function to get field info from python signature parameters"""
    text_info = func.__doc__

    for p in Sig(func).params:
        if p.annotation is not empty_param_attr:
            if p.default is not empty_param_attr:
                yield p.name, (p.annotation, Field(p.default))
            else:
                yield p.name, (p.annotation, Field(...))
        else:  # no annotations
            if p.default is not empty_param_attr:
                yield p.name, Field(p.default)
            else:
                yield p.name, (dflt_type, Field(...))


if __name__ == '__main__':
    from streamlitfront.base import dispatch_funcs
    from pydantic import BaseModel, Field, ValidationError, parse_obj_as

    # from streamlitfront.page_funcs import SimplePageFuncPydanticWrite

    # class ChoiceModel(BaseModel):
    #    single_selection: mk_choices_from_store(salary_store) = Field(
    #        ..., description="Only select a single item from a set."
    #    )

    configs = {'page_factory': SimplePageFuncPydanticWrite}

    app = dispatch_funcs([func], configs=configs)
    app()



import os
from extrude import mk_api, run_api
from py2http.decorators import handle_form_req, send_binary_resp
from i2 import Sig

from plunk.vf.extrude.edge_impulse_like import tag_wf, get_tagged_wf, mall


@handle_form_req
@Sig(tag_wf)
def tag_wf_input_handler(**kwargs):
    return kwargs


@send_binary_resp
def get_tagged_wf_output_handler(output, **input_kwargs):
    return output


handlers = [
    dict(endpoint=tag_wf, input_mapper=tag_wf_input_handler,),
    dict(endpoint=get_tagged_wf, output_mapper=get_tagged_wf_output_handler,)
    # get_tagged_wf
]
_app = mk_api(handlers, mall)
api_url = _app.openapi_spec['servers'][0]['url']


if __name__ == '__main__':
    run_api(_app)



"""
An app that loads either a wav file from local folder or records a sound
and visualizes the resulting numpy array 
"""
from typing import Mapping
from know.boxes import *
from functools import partial
from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY

from front.crude import Crudifier
import numpy as np

from streamlitfront import mk_app, binder as b
from streamlitfront.elements import SuccessNotification
from streamlitfront.elements import FileUploader

import soundfile as sf
from io import BytesIO

from plunk.sb.front_demo.user_story1.utils.tools import WaveForm
from typing import List


def mk_pipeline_maker_app_with_mall(
    mall: Mapping,
    *,
    step_factories: str = 'step_factories',
    steps: str = 'steps',
    steps_store=None,
    pipelines: str = 'pipelines',
    pipelines_store=None,
    data: str = 'data',
    data_store=None,
    learned_models=None,
    models_scores=None,
):
    if not b.mall():
        b.mall = mall
    mall = b.mall()

    crudifier = partial(Crudifier, mall=mall)

    @crudifier(output_store='sound_output')
    def upload_sound(train_audio: List[WaveForm], tag: str):

        return train_audio, tag

    config = {
        APP_KEY: {'title': 'Data Preparation'},
        RENDERING_KEY: {
            'upload_sound': {
                # NAME_KEY: "Data Loader",
                'execution': {
                    'inputs': {
                        'train_audio': {
                            ELEMENT_KEY: FileUploader,
                            'type': 'wav',
                            'accept_multiple_files': True,
                        },
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'Wav files loaded successfully.',
                    },
                },
            },
        },
    }

    funcs = [
        upload_sound,
    ]
    app = mk_app(funcs, config=config)

    return app


mall = dict(
    # Factory Input Stores
    sound_output=dict(),
    # Output Store
    data=dict(),
)


if __name__ == '__main__':

    app = mk_pipeline_maker_app_with_mall(mall,)

    app()



"""Utils for buildup"""
import os
from typing import Any
from sklearn.preprocessing import MinMaxScaler, StandardScaler
from sklearn.decomposition import PCA
import numpy as np
from dol.filesys import mk_tmp_dol_dir

# Being lazy with the definition of these:
FVs = Any
FittedModel = Any
Learner = Any


def get_a_root_directory_for_module_and_mk_tmp_dir_for_it(module_path, verbose=True):
    # if not (isinstance(module_path, str) and os.path.isdir(module_path)):
    #     module_path = module_path.__file__  # assuming it's an object with a __file__

    this_filename, *_ = os.path.splitext(module_path)
    this_filename = os.path.basename(this_filename)
    rootdir = mk_tmp_dol_dir(this_filename)
    if verbose:
        print(f'\n****************************************************')
        print(f'A temp directory was made for {this_filename}')
        print(f'The data will be saved here: {rootdir}')
        print(f'****************************************************')
    return rootdir


# ---------------------------------------------------------------------------------------
# The function(ality) we want to dispatch:
# Note: apply_model and learn_model both have the same structure, just different
#  parameters (names, annotations, defaults).
# TODO: apply_model and learn_model both have the same structure, just different
#  parameters (names, annotations, defaults), so...
#  Could make them from i2.wrapper tools. Should is a different question though.


def learn_model(learner: Learner, fvs: FVs, method='fit'):
    method_func = getattr(learner, method)
    return method_func(list(fvs))


def apply_model(fitted_model: FittedModel, fvs: FVs, method='transform'):
    method_func = getattr(fitted_model, method)
    # TODO: Should remove .tolist() (and possibly the list of list(fvs)).
    #  Not concern here.
    return method_func(list(fvs)).tolist()


# TODO: Make a sklearn-free and numpy-free version?
# class MinMaxScaler:
#     def fit(self, X):
#         self.min_ = min(X)

# ---------------------------------------------------------------------------------------
# The stores that will be used -- here, all stores are just dictionaries, but the
# contract is with the typing.Mapping (read-only here) interface.
# As we grow up, we'll use other mappings, such as:
# - server side RAM (as done here, simply)
# - server side persistence (files or any DB or file system thanks to the dol package)
# - computation (when you want the request for a key to actually launch a process that
#   will generate the value for you (some say you should be obvious to that detail))
# - client side RAM (when we figure that out)

# really, should be a str from a list of options, given by list(fvs_store)
FVsKey = str
# really, should be a str from a list of options, given by list(fitted_model_store)
FittedModelKey = str
Result = Any
ResultKey = str

mall = dict(
    fvs=dict(  # Mapping[FVsKey, FVs]
        train_fvs_1=np.array([[1], [2], [3], [5], [4], [2], [1], [4], [3]]),
        train_fvs_2=np.array([[1], [10], [5], [3], [4]]),
        test_fvs=np.array([[1], [5], [3], [10], [-5]]),
    ),
    fitted_model=dict(  # Mapping[FittedModelKey, FittedModel]
        fitted_model_1=MinMaxScaler().fit(
            [[1], [2], [3], [5], [4], [2], [1], [4], [3]]
        ),
        fitted_model_2=MinMaxScaler().fit([[1], [10], [5], [3], [4]]),
    ),
    model_results=dict(),  # Mapping[ResultKey, Result]
)

mall_with_learners = dict(
    mall,
    **dict(
        learner_store=dict(
            MinMaxScaler=MinMaxScaler(), StandardScaler=StandardScaler(), PCA=PCA(),
        )
    ),
)


def test_dispatch(dispatcher):
    w_apply_model = dispatcher(
        apply_model,
        param_to_mall_map=['fvs', 'fitted_model'],
        mall=mall,
        include_stores_attribute=True,
    )
    assert (
        w_apply_model('fitted_model_1', 'test_fvs')
        == [[0.0], [1.0], [0.5], [2.25], [-1.5]]
        == apply_model(
            fitted_model=w_apply_model.store_for_param['fitted_model'][
                'fitted_model_1'
            ],
            fvs=w_apply_model.store_for_param['fvs']['test_fvs'],
        )
    )



import pandas as pd
import plotly.express as px
import streamlit as st
import numpy as np
from utils import discretize, mk_audio_stream
import time

st.set_page_config(
    page_title='Stream a dataframe', layout='wide',
)

# create live stream from microphone
stream = mk_audio_stream()


# dashboard title
st.title('Stream a dataframe')


# creating a single-element container
placeholder = st.empty()

# live data
while True:
    time.sleep(0.1)

    data = np.frombuffer(stream.read(1024, exception_on_overflow=False), dtype=np.int16)

    dataL = discretize(data[0::2], func=np.std)

    d = {'sound': dataL}
    with placeholder.container():
        # fig = px.line(
        #     data_frame=pd.DataFrame(d),
        #     x=range(200),
        #     y="sound",
        #     range_y=[0, 1000],
        # )
        # st.write(fig)
        st.bar_chart(dataL)



"""
An app to take care of the initial 'sourcing' part of the data prep of audio ML
"""
from typing import Mapping
from know.boxes import *
from functools import partial
from typing import Callable, Iterable
from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY
from i2 import Pipe, Sig
from front.crude import Crudifier, prepare_for_crude_dispatch
from lined import LineParametrized

from streamlitfront import mk_app, binder as b
from streamlitfront.elements import (
    SelectBox,
    SuccessNotification,
    KwargsInput,
    PipelineMaker,
)
from streamlitfront.elements import (
    AudioRecorder,
    FileUploader,
    MultiSourceInput,
)
from streamlitfront.examples.util import Graph
import streamlit as st
from dataclasses import dataclass
from front.elements import OutputBase
from plunk.sb.front_experiments.streamlitfront_dataprep.data_prep2 import (
    # DFLT_WF_PATH,
    # DFLT_ANNOT_PATH,
    data_from_wav_folder,
)
import soundfile as sf
from io import BytesIO
import matplotlib.pyplot as plt

WaveForm = Iterable[int]


@dataclass
class AudioDisplay(OutputBase):
    def render(self):
        sound, tag = self.output
        if not isinstance(sound, str):
            sound = sound.getvalue()

        arr = sf.read(BytesIO(sound), dtype='int16')[0]
        tab1, tab2 = st.tabs(['Audio Player', 'Waveform'])
        with tab1:
            st.audio(sound)
        with tab2:
            fig, ax = plt.subplots(figsize=(15, 5))
            ax.plot(arr, label=f'Tag={tag}')
            ax.legend()
            st.pyplot(fig)
            # st.write(arr[:10])


def chunker(it, chk_size: int):
    return fixed_step_chunker(it=it, chk_size=chk_size, chk_step=chk_size)


@dataclass
class GraphOutput(OutputBase):
    use_container_width: bool = False

    def render(self):
        # with st.expander(self.name, True): #cannot nest expanders
        dag = self.output
        st.graphviz_chart(
            figure_or_dot=dag.dot_digraph(),
            use_container_width=self.use_container_width,
        )


def mk_pipeline_maker_app_with_mall(
    mall: Mapping,
    *,
    step_factories: str = 'step_factories',
    steps: str = 'steps',
    steps_store=None,
    pipelines: str = 'pipelines',
    pipelines_store=None,
    data: str = 'data',
    data_output=None,
    data_store=None,
):
    # TODO: Like to not have this binder logic involving streamlit state here! Contain it!
    if not b.mall():
        # TODO: Maybe it's here that we need to use know.malls.mk_mall?
        b.mall = mall
    mall = b.mall()
    if not b.selected_step_factory():
        b.selected_step_factory = 'data_loader'  # TODO make this dynamic

    crudifier = partial(Crudifier, mall=mall)

    steps_store = steps_store or steps
    data_store = data_store or data
    pipelines_store = pipelines_store or pipelines

    @crudifier(
        param_to_mall_map=dict(step_factory=step_factories), output_store=steps_store
    )
    def mk_step(step_factory: Callable, kwargs: dict):
        return partial(step_factory, **kwargs)

    #
    @crudifier(
        # TODO: Want to be able to do this and this only to have the effect
        # param_to_mall_map=dict(steps=steps),
        output_store=pipelines_store
    )
    def mk_pipeline(steps: Iterable[Callable]):
        return LineParametrized(*steps)

    @crudifier(
        # TODO: Does this work if pipelines_store is a mapping instead of a string?
        param_to_mall_map=dict(pipeline=pipelines_store, tagged_data='sound_output'),
        output_store='exec_outputs',
    )
    def exec_pipeline(pipeline: Callable, tagged_data):
        sound, tag = tagged_data
        if not isinstance(sound, str):
            sound = sound.getvalue()

        arr = sf.read(BytesIO(sound), dtype='int16')[0]
        result = list(
            pipeline(arr)()
        )  # TODO: because we use FuncFactories we need that hack
        st.write(result)
        return result

    @crudifier(
        # TODO: Does this work if pipelines_store is a mapping instead of a string?
        param_to_mall_map=dict(pipeline=pipelines_store),
        # output_store='exec_outputs'
    )
    def visualize_pipeline(pipeline: LineParametrized):

        return pipeline

    @crudifier(output_store='sound_output')
    def upload_sound(train_audio: WaveForm, tag: str):
        # mall["tag_store"] = tag
        return (train_audio, tag)

    @crudifier(param_to_mall_map={'result': 'sound_output'})
    def display_tag_sound(result):
        return result

    @crudifier(
        # TODO: Does this work if pipelines_store is a mapping instead of a string?
        param_to_mall_map=dict(step_factory=step_factories),
        output_store=data_store
        # output_store='exec_outputs'
    )
    def load_data(step_factory: Callable, kwargs: dict):
        st.write(mall)
        # result = partial(step_factory, **kwargs)
        result = step_factory(**kwargs)
        return result

    def get_step_name(step):
        return [k for k, v in mall[steps].items() if v == step][0]

    def get_selected_pipeline_sig():
        if not b.selected_pipeline():
            return Sig()
        return Sig(mall[pipelines][b.selected_pipeline()])

    config = {
        APP_KEY: {'title': 'Data Preparation'},
        RENDERING_KEY: {
            'upload_sound': {
                # NAME_KEY: "Data Loader",
                # "description": {"content": "A very simple learn model example."},
                'execution': {
                    'inputs': {
                        'train_audio': {
                            ELEMENT_KEY: MultiSourceInput,
                            'From a file': {ELEMENT_KEY: FileUploader, 'type': 'wav',},
                            'From the microphone': {ELEMENT_KEY: AudioRecorder},
                        },
                        # "tag": {
                        #     ELEMENT_KEY: TextInput,
                        # },
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'Wav loaded successfully.',
                    },
                },
            },
            'display_tag_sound': {
                'execution': {
                    'inputs': {
                        'result': {
                            ELEMENT_KEY: SelectBox,
                            'options': mall['sound_output'],
                        },
                    },
                    'output': {ELEMENT_KEY: AudioDisplay,},
                },
            },
            'load_data': {
                NAME_KEY: 'Data Loader',
                'execution': {
                    'inputs': {
                        'step_factory': {'value': b.selected_step_factory,},
                        'kwargs': {
                            'func_sig': Sig(
                                mall[step_factories][b.selected_step_factory()]
                            ),
                        },
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'The step has been created successfully.',
                    },
                },
            },
            'mk_step': {
                NAME_KEY: 'Pipeline Step Maker',
                'execution': {
                    'inputs': {
                        'step_factory': {'value': b.selected_step_factory,},
                        'kwargs': {
                            'func_sig': Sig(
                                mall[step_factories][b.selected_step_factory()]
                            ),
                        },
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'The step has been created successfully.',
                    },
                },
            },
            'mk_pipeline': {
                NAME_KEY: 'Pipeline Maker',
                'execution': {
                    'inputs': {
                        steps: {
                            ELEMENT_KEY: PipelineMaker,
                            'items': list(mall[steps].values()),
                            'serializer': get_step_name,
                        },
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'The pipeline has been created successfully.',
                    },
                },
            },
            'exec_pipeline': {
                NAME_KEY: 'Pipeline Executor',
                'execution': {
                    'inputs': {
                        'pipeline': {'value': b.selected_pipeline,},
                        'data': {
                            ELEMENT_KEY: SelectBox,
                            'options': mall['sound_output'],
                        },
                    }
                },
            },
            'visualize_pipeline': {
                NAME_KEY: 'Pipeline Visualization',
                'execution': {
                    'inputs': {'pipeline': {'value': b.selected_pipeline,},},
                    'output': {
                        ELEMENT_KEY: GraphOutput,
                        NAME_KEY: 'Flow',
                        'use_container_width': True,
                    },
                },
            },
        },
    }

    funcs = [
        upload_sound,
        display_tag_sound,
        # load_data,
        mk_step,
        mk_pipeline,
        exec_pipeline,
        visualize_pipeline,
    ]
    app = mk_app(funcs, config=config)

    return app


if __name__ == '__main__':

    mall = dict(
        # Factory Input Stores
        sound_output=dict(),
        step_factories=dict(),
        # Output Store
        data=dict(),
        steps=dict(),
        pipelines=dict(),
        exec_outputs=dict(),
    )

    crudifier = partial(prepare_for_crude_dispatch, mall=mall)

    step_factories = dict(
        # Source Readers
        data_loader_factory=FuncFactory(data_from_wav_folder),
        data_loader=data_from_wav_folder,
        # Chunkers
        chunker=FuncFactory(chunker),
    )

    mall['step_factories'] = step_factories

    app = mk_pipeline_maker_app_with_mall(
        mall, step_factories='step_factories', steps='steps', pipelines='pipelines'
    )

    app()



if __name__ == '__main__':
    # What I want to dispatch
    from plunk.tw.py2py_front_example.simple_pycode import foo, bar, confuser

    funcs = [foo, bar, confuser]

    # My dispatching
    from py2http import run_app

    run_app(funcs, publish_openapi=True)



"""
Same as take_04_model_run, but where the dispatch is not as manual.
"""
from front.crude import KT, StoreName, Mall
from crude.extrude_crude.extrude_crude_util import mall, np, apply_model

# ---------------------------------------------------------------------------------------
# dispatchable function:
from front.crude import prepare_for_crude_dispatch

f = prepare_for_crude_dispatch(apply_model, param_to_mall_map=mall)
assert all(
    f('fitted_model_1', 'test_fvs') == np.array([[0.0], [1.0], [0.5], [2.25], [-1.5]])
)


def simple_mall_dispatch_core_func(
    key: KT, action: str, store_name: StoreName, mall: Mall
):
    if not store_name:
        # if store_name empty, list the store names (i.e. the mall keys)
        return list(mall)
    else:  # if not, get the store
        store = mall[store_name]

    if action == 'list':
        key = key.strip()  # to handle some invisible whitespace that would screw things
        return list(filter(lambda k: key in k, store))
    elif action == 'get':
        return store[key]


# TODO: the function doesn't see updates made to mall. Fix.
# Just the partial (with mall set), but without mall arg visible (or will be dispatched)
def explore_mall(key: KT, action: str, store_name: StoreName):
    return simple_mall_dispatch_core_func(key, action, store_name, mall=mall)


# Attempt to do this wit i2.wrapper
# from functools import partial
# from i2.wrapper import rm_params_ingress_factory, wrap
#
# without_mall_param = partial(
#     wrap, ingress=partial(rm_params_ingress_factory, params_to_remove="mall")
# )
# mall_exploration_func = without_mall_param(
#     partial(simple_mall_dispatch_core_func, mall=mall)
# )
# mall_exploration_func.__name__ = "explore_mall"

if __name__ == '__main__':
    from crude.util import ignore_import_problems

    with ignore_import_problems:
        from streamlitfront.base import dispatch_funcs
        from functools import partial

        dispatchable_apply_model = prepare_for_crude_dispatch(
            apply_model, output_store='model_results'
        )
        # extra, to get some defaults in:
        dispatchable_apply_model = partial(
            dispatchable_apply_model, fitted_model='fitted_model_1', fvs='test_fvs',
        )
        app = dispatch_funcs([dispatchable_apply_model, explore_mall])
        app()



# from tested import validate_codec
#
#
# Empty1 = type('Empty', (), {})()
# print("\nEmpty1 = type('Empty', (), {})()")
# print(f'{validate_codec(Empty1)=}')
#
#
# _Empty2 = type('Empty', (), {})
# _Empty2.__module__ = __name__
# Empty2 = _Empty2()
# print('\nsame, with _Empty2.__module__ = __name__; Empty2 = _Empty()')
# print(f'{validate_codec(Empty2)=}')
#
# _Empty3 = type('Empty', (), {})
# Empty3 = _Empty3()
# Empty3.__module__ = __name__
# print('\nwith Empty3.__module__ = __name__ (added to instance instead of class)')
# print(f'{validate_codec(Empty3)=}')
#
#
# class _Empty4:
#     ...
#
#
# Empty4 = _Empty4()
# print('\nDefining as ``class _Empty4: ...``')
# print(f'{validate_codec(Empty4)=}')
#
#
# class Empty5:
#     ...
#
#
# Empty5 = Empty5()  # overridding class (now no one else can use it!"
#
# print('\nOverwriding class Empty5 with instance Empty5')
# print(f'{validate_codec(Empty5)=}')
#
#
# class _Empty6:
#     def __repr__(self):
#         return f'Empty'
#
#
# Empty6 = _Empty6()
#
# print('\nLike Empty4, but with nice repr')
# print(f'{validate_codec(Empty6)=}')
#
# from i2 import mk_sentinel
#
# #
# #
# Empty9 = mk_sentinel('Empty9')
#
#
# print('\nWith boltons.typeutils.Sentinel')
# print(f'{validate_codec(Empty9)=}')
#
# import pickle
#
# # pickle.dumps(Empty9)


def foo():
    """
    >>> import pickle
    >>> class A: ...
    >>> unpickled_A = pickle.loads(pickle.dumps(A))
    >>> unpickled_A == A
    True
    """


if __name__ == '__main__':
    import doctest

    doctest.testmod()



"""
An app that loads either a wav file from local folder or records a sound
and visualizes the resulting numpy array 
"""
from functools import partial
from typing import List, Mapping, Iterable, Callable
from i2 import FuncFactory, Sig
from lined import LineParametrized

from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY
from front.crude import Crudifier
from streamlitfront import mk_app, binder as b
from streamlitfront.elements import (
    SelectBox,
    SuccessNotification,
    PipelineMaker,
)
from streamlitfront.elements import FileUploader

from olab.types import (
    Step,
    Pipeline,
    WaveForm,
    AudioArrayDisplay,
    GraphOutput,
    ArrayPlotter,
)
from olab.util import (
    simple_chunker,
    scores_to_intervals,
    simple_featurizer,
    ArrayWithIntervalsPlotter,
    clean_dict,
)

from olab.base import learn_outlier_model, apply_fitted_model


def mk_pipeline_maker_app_with_mall(
    mall: Mapping,
    *,
    step_factories: str = 'step_factories',
    steps: str = 'steps',
    steps_store=None,
    pipelines: str = 'pipelines',
    pipelines_store=None,
    data: str = 'data',
    data_store=None,
    learned_models=None,
    models_scores=None,
):

    if not b.mall():
        b.mall = mall
    mall = b.mall()
    if not b.selected_step_factory():
        b.selected_step_factory = 'chunker'  # TODO make this dynamic

    crudifier = partial(Crudifier, mall=mall)

    steps_store = steps_store or steps
    data_store = data_store or data
    pipelines_store = pipelines_store or pipelines

    @crudifier(
        param_to_mall_map=dict(step_factory=step_factories), output_store=steps_store
    )
    def mk_step(step_factory: Callable, kwargs: dict):
        kwargs = clean_dict(kwargs)  # TODO improve that logic
        step = partial(step_factory, **kwargs)()
        result = Step(step=step, step_factory=step_factory)
        return result

    @crudifier(
        param_to_mall_map=dict(step_to_modify=steps_store), output_store=steps_store
    )
    def modify_step(step_to_modify: Step, kwargs: dict):

        kwargs = clean_dict(kwargs)  # TODO improve that logic
        step_factory = step_to_modify.step_factory
        step = partial(step_factory, **kwargs)()
        return Step(step=step, step_factory=step_factory)

    @crudifier(output_store=pipelines_store,)
    def mk_pipeline(steps: Iterable[Callable]):
        named_funcs = [(get_step_name(step), step) for step in steps]
        pipeline = Pipeline(steps=steps, pipe=LineParametrized(*named_funcs))
        return pipeline

    @crudifier(
        param_to_mall_map=dict(pipeline=pipelines_store), output_store=pipelines_store,
    )
    def modify_pipeline(pipeline, steps):
        named_funcs = [(get_step_name(step), step) for step in steps]
        pipe = LineParametrized(*named_funcs)
        return Pipeline(steps=named_funcs, pipe=pipe)

    learn_outlier_model_crudified = crudifier(
        param_to_mall_map=dict(
            tagged_data='sound_output', preprocess_pipeline='pipelines'
        ),
        output_store='learned_models',
    )(learn_outlier_model)

    apply_fitted_model_crudified = crudifier(
        param_to_mall_map=dict(
            tagged_data='sound_output',
            preprocess_pipeline='pipelines',
            fitted_model='learned_models',
        ),
        output_store='models_scores',
    )(apply_fitted_model)

    @crudifier(param_to_mall_map=dict(pipeline=pipelines_store),)
    def visualize_pipeline(pipeline: Pipeline):

        return pipeline

    @crudifier(param_to_mall_map=dict(scores='models_scores'),)
    def visualize_scores(scores, threshold=80, num_segs=3):

        intervals = scores_to_intervals(scores, threshold, num_segs)

        return scores, intervals

    @crudifier(output_store='sound_output')
    def upload_sound(train_audio: List[WaveForm], tag: str):
        return train_audio, tag

    def get_step_name(step):
        return [k for k, v in mall[steps].items() if v.step == step][0]

    def get_selected_step_factory_sig():
        selected_step_factory = mall['step_factories'].get(
            b.selected_step_factory.get()
        )
        if selected_step_factory:
            return Sig(selected_step_factory)

    def get_step_to_modify_factory_sig():
        selected_step_factory = (
            mall['steps'].get(b.selected_step_to_modify.get()).step_factory
        )
        if selected_step_factory:
            return Sig(selected_step_factory)

    def on_select_pipeline(pipeline):
        b.steps_of_selected_pipeline.set(mall['pipelines'][pipeline].steps)

    config = {
        APP_KEY: {'title': 'Data Preparation'},
        RENDERING_KEY: {
            'upload_sound': {
                # NAME_KEY: "Data Loader",
                'execution': {
                    'inputs': {
                        'train_audio': {
                            ELEMENT_KEY: FileUploader,
                            'type': 'wav',
                            'accept_multiple_files': True,
                        },
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'Wav files loaded successfully.',
                    },
                },
            },
            'display_tag_sound': {
                'execution': {'output': {ELEMENT_KEY: AudioArrayDisplay,},},
            },
            'mk_step': {
                NAME_KEY: 'Pipeline Step Maker',
                'execution': {
                    'inputs': {
                        'step_factory': {'value': b.selected_step_factory,},
                        'kwargs': {'func_sig': get_selected_step_factory_sig},
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'The step has been created successfully.',
                    },
                },
            },
            'modify_step': {
                NAME_KEY: 'Modify Step',
                'execution': {
                    'inputs': {
                        'step_to_modify': {'value': b.selected_step_to_modify,},
                        'kwargs': {'func_sig': get_step_to_modify_factory_sig},
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'The step has been created successfully.',
                    },
                },
            },
            'mk_pipeline': {
                NAME_KEY: 'Pipeline Maker',
                'execution': {
                    'inputs': {
                        steps: {
                            ELEMENT_KEY: PipelineMaker,
                            'items': [v.step for v in mall[steps].values()],
                            'serializer': get_step_name,
                        },
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'The pipeline has been created successfully.',
                    },
                },
            },
            'modify_pipeline': {
                NAME_KEY: 'Pipeline Modify',
                'execution': {
                    'inputs': {
                        'pipeline': {
                            ELEMENT_KEY: SelectBox,
                            'value': b.selected_pipeline,
                            'on_value_change': on_select_pipeline,
                        },
                        steps: {
                            ELEMENT_KEY: PipelineMaker,
                            'items': [v.step for v in mall[steps].values()],
                            'steps': b.steps_of_selected_pipeline(),
                            'serializer': get_step_name,
                        },
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'The pipeline has been modified successfully.',
                    },
                },
            },
            'visualize_pipeline': {
                NAME_KEY: 'Pipeline Visualization',
                'execution': {
                    'inputs': {'pipeline': {'value': b.selected_pipeline,},},
                    'output': {
                        ELEMENT_KEY: GraphOutput,
                        NAME_KEY: 'Flow',
                        'use_container_width': True,
                    },
                },
            },
            'visualize_scores': {
                NAME_KEY: 'Scores Visualization',
                'execution': {'output': {ELEMENT_KEY: ArrayWithIntervalsPlotter,},},
            },
            'simple_model': {
                NAME_KEY: 'Learn model',
                'execution': {'output': {ELEMENT_KEY: ArrayPlotter,},},
            },
            'apply_fitted_model': {
                NAME_KEY: 'Apply model',
                'execution': {'output': {ELEMENT_KEY: ArrayPlotter,},},
            },
        },
    }

    funcs = [
        upload_sound,
        mk_step,
        modify_step,
        mk_pipeline,
        modify_pipeline,
        learn_outlier_model_crudified,
        apply_fitted_model_crudified,
        visualize_pipeline,
        visualize_scores,
    ]
    app = mk_app(funcs, config=config)

    return app


mall = dict(
    # Factory Input Stores
    sound_output=dict(),
    step_factories=dict(
        # ML
        chunker=FuncFactory(simple_chunker),
        featurizer=FuncFactory(simple_featurizer),
    ),
    # Output Store
    data=dict(),
    steps=dict(),
    pipelines=dict(),
    exec_outputs=dict(),
    learned_models=dict(),
    models_scores=dict(),
)


if __name__ == '__main__':

    app = mk_pipeline_maker_app_with_mall(
        mall, step_factories='step_factories', steps='steps', pipelines='pipelines'
    )

    app()



from meshed import code_to_dag
from plunk.sb.front_experiments.data_visualizer.utils.tools import (
    store_to_key_fvs,
    key_fvs_to_tag_fvs,
    mk_Xy,
)

d_input = {
    'store_to_key_fvs': store_to_key_fvs,
    'key_fvs_to_tag_fvs': key_fvs_to_tag_fvs,
    'mk_Xy': mk_Xy,
}


@code_to_dag(func_src=d_input)
def load_dataset():
    key_fvs = store_to_key_fvs(wf_store)
    tag_fv_iterator = key_fvs_to_tag_fvs(key_fvs, annots_df=annots_df)
    X, y = mk_Xy(tag_fv_iterator)



import pytest
from http2py.client import HttpClient
from selenium.webdriver import Chrome, ChromeOptions
from webdriver_manager.chrome import ChromeDriverManager
from streamlitfront.tests.common import select_func, send_input, compute_output
from time import sleep

DFLT_VAX = 0.8


def _factor(vax, vax_factor):
    return (
        # vax is proportion of vaxinated and vax_factor how much less
        # (compared to non-vax) a given probability would be for vaxinated):
        vax * vax_factor
        +
        # (1 - vax) is the proportion of non vaxinated folks and 1.0 means "no change"
        (1 - vax) * 1.0
    )


def r(exposed=6, infect_if_expose=1 / 5):
    return exposed * infect_if_expose


def infected(r=1.2, vax=0.8, infection_vax_factor=0.15):
    return r * _factor(vax, infection_vax_factor)


def die(
    infected, die_if_infected=0.05, vax=DFLT_VAX, death_vax_factor=0.05,
):
    return infected * die_if_infected * _factor(vax, death_vax_factor)


def death_toll(die, population=1e6):
    return int(die * population)


@pytest.mark.parametrize(
    'call_specs',
    [
        (
            [
                dict(
                    func=r,
                    inputs=dict(exposed=10, infect_if_expose=0.3,),
                    expected_result=3.0,
                ),
                dict(
                    func=infected,
                    inputs=dict(r=2, vax=0.5, infection_vax_factor=0.5),
                    expected_result=1.5,
                ),
                dict(
                    func=die,
                    inputs=dict(
                        infected=1.5,
                        die_if_infected=0.25,
                        vax=0.5,
                        death_vax_factor=0.25,
                    ),
                    expected_result=0.234375,
                ),
                dict(
                    func=death_toll,
                    inputs=dict(die=0.234375, population=2000000,),
                    expected_result=468750.0,
                ),
            ]
        )
    ],
)
def test_published_app(call_specs, surface='ui'):
    def parse_spec(spec):
        return (
            spec['func'],
            spec['inputs'],
            spec['expected_result'],
        )

    def test_api():
        print(f'Validating the API...')
        api = HttpClient(url='https://extrudeploy-api.herokuapp.com/openapi')
        for spec in call_specs:
            func, inputs, expected_result = parse_spec(spec)
            endpoint = getattr(api, func.__name__)
            api_result = endpoint(**inputs)
            print(f'Asserting that {api_result=} == {expected_result=}')
            assert api_result == expected_result
            sleep(1)
        print(f'All good with the API!')
        sleep(1)

    def test_ui():
        print(f'Validating the UI...')

        def call_ui(idx, func, inputs):
            select_func(idx, dom)
            for idx, value in enumerate(inputs.values()):
                send_input(value, idx, dom)
                sleep(1)
            return compute_output(func, dom)

        options = ChromeOptions()
        options.add_argument('--window-size=1920,1080')
        dom = Chrome(ChromeDriverManager().install(), options=options)
        dom.get('https://extrudeploy.herokuapp.com')
        try:
            for idx, spec in enumerate(call_specs):
                func, inputs, expected_result = parse_spec(spec)
                ui_result = call_ui(idx, func, inputs)
                print(f'Asserting that {ui_result=} == {expected_result=}')
                assert float(ui_result) == expected_result
                # assert ui_result == expected_result
            print(f'All good with the UI!')
            sleep(1)
        finally:
            dom.close()

    if surface == 'ui':
        test_ui()
    elif surface:
        test_api()



from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY
from collections.abc import Callable
from front.crude import Crudifier

from streamlitfront import mk_app, binder as b
from streamlitfront.elements import SelectBox
from typing import Any
from streamlitfront.elements import MultiSourceInput
import streamlit as st
from front.elements import InputBase
from i2 import Sig
from dataclasses import dataclass
from streamlitfront.elements import TextInput, SelectBox, ExecSection, TextOutput
from functools import partial
from front.crude import Crudifier
from streamlitfront.data_binding import BoundData
from i2.deco import FuncFactory
from plunk.sb.front_experiments.streamlitfront_dataprep.data_prep2 import (
    # DFLT_WF_PATH,
    # DFLT_ANNOT_PATH,
    data_from_wav_folder,
    data_from_csv,
    store_to_key_fvs,
    key_fvs_to_tag_fvs,
    mk_Xy,
)
from odat.mdat.vacuum import DFLT_CHUNKER, DFLT_FEATURIZER
from typing import Optional
from slang.chunkers import fixed_step_chunker
from front.base import prepare_for_dispatch
from front.crude import simple_mall_dispatch_core_func
from front.util import inject_enum_annotations
from front.base import prepare_for_dispatch
from front.crude import KT, StoreName


# # ============ MALL ===============
def make_chunker(chk_size: int, chk_step: Optional[int] = None):
    return partial(fixed_step_chunker, chk_size=chk_size, chk_step=chk_step)


if not b.mall():
    b.mall = {
        'func_store': {},
        'featurizer': {},
        'featurizer_cls': {},
        'chunker': {},
        'factories': {'chunker': make_chunker},
        'wf_store': {},
        'wf_store_cls': {'data_from_wav_folder': data_from_wav_folder},
        'annots_loader': {},
        'annots_loader_cls': {'data_from_csv': data_from_csv},
    }

mall = b.mall()
chunker = prepare_for_dispatch(make_chunker, output_store=mall['chunker'],)


DFLT_CHUNKER_MAKER = lambda: DFLT_CHUNKER
DFLT_FEATURIZER_MAKER = lambda: DFLT_FEATURIZER
DFLT_WF_PATH = '/Users/sylvain/Dropbox/Otosense/VacuumEdgeImpulse/'
DFLT_ANNOT_PATH = '/Users/sylvain/Dropbox/sipyb/Testing/data/annots_vacuum.csv'


FixedSizeChunker = DFLT_CHUNKER
featurizer = DFLT_FEATURIZER

# def data_loader(**kwargs):
#     return data_from_wav_folder(filepath)

# DFLT_CHUNKER_MAKER = lambda: DFLT_CHUNKER
# DFLT_FEATURIZER_MAKER = lambda: DFLT_FEATURIZER
# DFLT_WF_PATH = "/Users/sylvain/Dropbox/Otosense/VacuumEdgeImpulse/"
# DFLT_ANNOT_PATH = "/Users/sylvain/Dropbox/sipyb/Testing/data/annots_vacuum.csv"

# metadata = {
#     "FixedSizeChunker": {"func": DFLT_CHUNKER, "out": "chks"},
#     # 'FixedSizeChunker100': {'out': 'chks'},
#     # 'ThresholdChunker': {'out': 'chks'},
#     "FixedSizeChunkerMaker": {"func": DFLT_CHUNKER_MAKER, "out": "chunker"},
#     "FeaturizerMaker": {"func": DFLT_FEATURIZER_MAKER, "out": "featurizer"},
#     # 'key_fvs_to_tag_fvs': {'func': key_fvs_to_tag_fvs},
#     # 'Featurizer': {'func': DFLT_FEATURIZER, 'out': 'fvs'},
#     "store_to_key_fvs": {"func": store_to_key_fvs, "out": "key_fvs"},
#     "key_fvs_to_tag_fvs": {"func": key_fvs_to_tag_fvs, "out": "tag_fv_iterator"},
#     "WfStoreMaker": {
#         "func": data_from_wav_folder,
#         "out": "wf_store",
#         "bind": {"filepath": "wf_filepath"},
#     },
#     "AnnotsStoreMaker": {"func": data_from_csv, "out": "annots_df"},
#     "mk_Xy": {"func": mk_Xy},
# }

from meshed import DAG
from sklearn.preprocessing import MinMaxScaler


from streamlitfront.elements import SelectBox


from front.spec_maker_base import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY
from streamlitfront.base import mk_app


# crudifier_output = Crudifier(output_store='func_store', mall=mall)

# @Crudifier(output_store='func_store', mall = mall)


def dag_from_mall():
    pass


def select_func(func, store_name, kwargs):
    f = metadata[func]
    result_func = partial(f, **kwargs)
    mall[store_name][func] = result_func
    sig = Sig(result_func)
    st.write(f'added function with sig={sig}')
    return result_func


# @inject_enum_annotations(action=["list", "get"], store_name=mall)


def convert(arg):
    return None


# @inject_enum_annotations(action=["list", "get"], store_name=mall)
def explore_mall(
    store_name: StoreName, key: KT, action: str,
):
    args = [key, action, store_name]
    [key, action, store_name] = list(map(convert, args))

    return simple_mall_dispatch_core_func(key, action, store_name, mall=mall)


def get_kwargs(**kwargs):
    return kwargs


def populate_list_funcs():
    value = b.selected_func()
    b.list_funcs().append(metadata[value])


if not b.selected_func():
    b.selected_func = 'chunker'

if not b.list_funcs():
    b.list_funcs = [select_func]

data = ['chunker', 'featurizer', 'data_loader']
metadata = {
    'chunker': chunker,
    'featurizer': featurizer,
    'data_loader': data_from_wav_folder,
    'csv_loader': data_from_csv,
}


@dataclass
class KwargsInput(InputBase):
    func_sig: Sig = None

    def __post_init__(self):
        super().__post_init__()
        self.get_kwargs = self.func_sig(get_kwargs)
        self.inputs = self._build_inputs_from_sig()

    def render(self):
        exec_section = ExecSection(
            obj=self.get_kwargs,
            inputs=self.inputs,
            output={ELEMENT_KEY: TextOutput},
            auto_submit=True,
            on_submit=self._return_kwargs,
            use_expander=False,
        )
        exec_section()
        return self.value()

    def _build_inputs_from_sig(self):
        return {
            name: {ELEMENT_KEY: TextInput, 'bound_data_factory': BoundData}
            for name in self.func_sig
        }

    def _return_kwargs(self, output):
        self.value.set(output)


def send_message(output):
    b.update_store()


from typing import Tuple


def dummy(t: Tuple[int, int]):
    return t


if __name__ == '__main__':
    app = mk_app(
        [select_func, explore_mall],
        config={
            APP_KEY: {'title': 'Rendering map'},
            RENDERING_KEY: {
                'select_func': {
                    'execution': {
                        'inputs': {
                            'func': {
                                ELEMENT_KEY: SelectBox,
                                'options': data,
                                'value': b.selected_func,
                                # "on_value_change": populate_list_funcs,
                            },
                            'store_name': {
                                ELEMENT_KEY: SelectBox,
                                'options': ['func_store'],
                                # "value": b.selected_func,
                                # "on_value_change": populate_list_funcs,
                            },
                            'kwargs': {
                                ELEMENT_KEY: KwargsInput,
                                'func_sig': Sig(metadata[b.selected_func()]),
                            },
                        },
                        'on_submit': send_message,
                    }
                },
            },
        },
    )
    app()
    st.write(mall)



from i2 import Sig
from meshed import DAG, FuncNode
from meshed.base import _func_nodes_to_graph_dict
from meshed.dag import _separate_func_nodes_and_var_nodes
from meshed.itools import topological_sort

# separate with comments lines
def pairs(xs):
    if len(xs) <= 1:
        return xs
    else:
        pairs = list(zip(xs, xs[1:]))
    return pairs


def mk_mock_funcnode(arg, out):
    @Sig(arg)
    def func():
        pass

    # name = "_mock_" + str(arg) + "_" + str(out)  # f-string
    name = f'_mock_{str(arg)}_{str(out)}'  # f-string

    return FuncNode(func=func, out=out, name=name)


def funcnodes_from_pairs(pairs):
    return list(map(mk_mock_funcnode_from_tuple, pairs))


def curry(func):
    def res(*args):
        return func(tuple(args))

    return res


def uncurry(func):
    def res(tup):
        return func(*tup)

    return res


mk_mock_funcnode_from_tuple = uncurry(mk_mock_funcnode)


def reorder_on_constraints(funcnodes, outs):
    extra_nodes = funcnodes_from_pairs(pairs(outs))
    funcnodes += extra_nodes
    graph = _func_nodes_to_graph_dict(funcnodes)
    nodes = topological_sort(graph)
    print('after ordering:', nodes)
    ordered_nodes = [node for node in nodes if node not in extra_nodes]
    func_nodes, var_nodes = _separate_func_nodes_and_var_nodes(ordered_nodes)

    return func_nodes, var_nodes



import streamlit as st


from front.spec_maker import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY
from streamlitfront.base import mk_app
from meshed import code_to_dag, DAG
from front.elements import FileUploaderBase
from streamlitfront.elements import implement_input_component
from typing import Iterable
from streamlitfront.examples.graph_component import Graph
from utils import (
    discretize,
    wav_complex_display,
    arr_from_bytes,
    bytes_from_uploaded,
)

WaveForm = Iterable[int]

WavUploader = implement_input_component(
    FileUploaderBase,
    component_factory=st.file_uploader,
    input_value_callback=lambda input, self: wav_complex_display(input),
)


@code_to_dag
def display_wf(wf: WaveForm):
    result = display_wf(wf)


config_ = {
    APP_KEY: {'title': 'Simple Load and Display'},
    RENDERING_KEY: {
        DAG: {
            'execution': {
                'inputs': {
                    'wf': {
                        # ELEMENT_KEY: FILE_UPLOADER_COMPONENT,
                        ELEMENT_KEY: WavUploader,
                        'type': 'wav',
                    },
                }
            },
            'graph': {ELEMENT_KEY: Graph, NAME_KEY: 'Flow',},
        }
    },
}

app = mk_app([display_wf], config=config_)
app()



import numpy as np
from typing import Iterable, Callable, List
from functools import partial
from plunk.sb.front_demo.user_story1.utils.tools import (
    WaveForm,
    Stroll,
    clean_dict,
    scores_to_intervals,
    # lined_dag,
    tagged_sounds_to_single_array,
    tagged_sound_to_array,
    assert_dims,
)
from meshed.makers import lined_dag
from omodel.outliers.pystroll import OutlierModel as Stroll
from meshed import DAG


def mk_step(step_factory: Callable, kwargs: dict):
    kwargs = clean_dict(kwargs)  # TODO improve that logic
    step = partial(step_factory, **kwargs)()

    return step


#


def mk_pipeline(steps: Iterable[Callable]):
    return lined_dag(steps)
    # return LineParametrized(*steps)


def exec_pipeline(pipeline: Callable, tagged_data):

    sound, _ = tagged_sound_to_array(*tagged_data)
    wfs = np.array(sound)
    wfs = assert_dims(wfs)

    result = pipeline(wfs)
    return result


def learn_outlier_model(tagged_data, preprocess_pipeline, n_centroids=5):
    sound, tag = tagged_sounds_to_single_array(*tagged_data)
    wfs = np.array(sound)

    wfs = assert_dims(wfs)

    fvs = preprocess_pipeline(wfs)
    model = Stroll(n_centroids=n_centroids)
    model.fit(X=fvs)

    return model


def apply_fitted_model(tagged_data, preprocess_pipeline, fitted_model):
    sound, tag = tagged_sounds_to_single_array(*tagged_data)
    wfs = np.array(sound)
    wfs = assert_dims(wfs)

    fvs = preprocess_pipeline(wfs)
    scores = fitted_model.score_samples(X=fvs)
    return scores


def visualize_pipeline(pipeline: DAG):

    return pipeline


def visualize_scores(scores, threshold=80, num_segs=3):

    intervals = scores_to_intervals(scores, threshold, num_segs)

    return scores, intervals


def upload_sound(train_audio: List[WaveForm], tag: str):
    # sound, tag = train_audio, tag
    # if not isinstance(sound, bytes):
    #     sound = sound.getvalue()

    # arr = sf.read(BytesIO(sound), dtype='int16')[0]
    # return arr, tag
    return train_audio, tag


def display_tag_sound(result):
    return result


def learn_apply_model(tagged_data, pipeline):
    n_centroids = 15
    sound, _ = tagged_sounds_to_single_array(*tagged_data)
    wfs = np.array(sound)
    wfs = assert_dims(wfs)

    fvs = pipeline(wfs)
    model = Stroll(n_centroids=n_centroids)
    fitted_model = model.fit(X=fvs)
    scores = fitted_model.score_samples(X=fvs)
    return scores



from functools import partial
from typing import Any
from front.crude import Crudifier

mall = dict(tagged_wf=dict())
crudifier = partial(Crudifier, mall=mall, include_stores_attribute=True)
WaveForm = Any


@crudifier(output_store='tagged_wf', output_trans=lambda: None)
def tag_wf(wf: WaveForm, tag: str):
    return (wf, tag)


@crudifier(param_to_mall_map=dict(x='tagged_wf'))
def get_tagged_wf(x: Any):
    return x


funcs = [tag_wf, get_tagged_wf]



from dataclasses import dataclass
from streamlitfront.elements import TextInput, FileUploader
from front.elements import InputBase
from hear import WavLocalFileStore
import pandas as pd
import soundfile as sf

from dol import FilesOfZip, wrap_kvs, filt_iter
from io import BytesIO


def my_obj_of_data(b):
    return sf.read(BytesIO(b), dtype='float32')[0]


# @wrap_kvs(obj_of_data=lambda b: sf.read(BytesIO(b), dtype="int16")[0])
@wrap_kvs(obj_of_data=my_obj_of_data)
@filt_iter(filt=lambda x: not x.startswith('__MACOSX') and x.endswith('.wav'))
class WfStore(FilesOfZip):
    """Waveform access. Keys are .wav filenames and values are numpy arrays of int16 waveform."""

    pass


@dataclass
class ZipWavDataLoader(FileUploader):
    # root_dir: str = None

    def __post_init__(self):
        # super().__post_init__()
        self.store = WfStore(self.output)

        # self.annots = pd.read_csv(self.annot_path)

    def render(self):
        print('done!')
        return len(self.store)


@dataclass
class DataLoader2(TextInput):
    # root_dir: str = None
    paths: str = None

    # def __post_init__(self):
    #     # super().__post_init__()
    #     # self.store = WavLocalFileStore(self.root_dir)

    #     # self.annots = pd.read_csv(self.annot_path)
    #     pass

    def render(self):
        print('done!')
        return len(self.output)



from dataclasses import dataclass
import streamlit as st
import numpy as np
import pandas as pd
from front.crude import Crudifier
import matplotlib.pyplot as plt
from typing import Any

# from plunk.sb.front_experiments.data_visualizer.dags.wfstore_and_annots_to_dataset import (
#     load_dataset,
# )
# from plunk.sb.front_experiments.data_visualizer.components.elements import (
#     # DataLoader,
#     # DataLoader2,
#     ZipWavDataLoader,
# )
# from functools import partial
from plunk.sb.front_experiments.data_visualizer.utils.tools import (
    store_to_key_fvs,
    key_fvs_to_tag_fvs,
    mk_Xy,
)
from streamlitfront import mk_app, binder as b
from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY
from streamlitfront.elements import SelectBox
from front.elements import OutputBase
import umap
import umap.plot
from hear import WavLocalFileStore


root_dir = '/Users/sylvain/Dropbox/_odata/sound/vacuum'
annots_path = '../data/annots_vacuum.csv'


# ============ BACKEND ============


def load_dataset(folder_path: str, annot_path: str):
    wf_store = WavLocalFileStore(folder_path)
    df = pd.read_csv(annot_path)
    key_fvs = store_to_key_fvs(wf_store)
    tag_fv_iterator = key_fvs_to_tag_fvs(key_fvs, annots_df=df)
    X, y = mk_Xy(tag_fv_iterator)  # Cmd+]

    return X, y


def plot_umap(Xy: Any):
    return Xy


# ============ END BACKEND ============


# ============ FRONT END   ============


if not b.mall():
    b.mall = dict(Xy={})
mall = b.mall()

load_dataset = Crudifier(output_store='Xy', mall=mall)(load_dataset)


plot_umap = Crudifier(param_to_mall_map=['Xy'], mall=mall)(plot_umap)


@dataclass
class UmapPlotter(OutputBase):
    def render(self):
        X, y = self.output
        mapper = umap.UMAP().fit(X)
        fig, ax = plt.subplots()
        show_legend = st.checkbox(label='Show legend')
        umap.plot.points(mapper, labels=np.array(y), show_legend=show_legend, ax=ax)
        st.pyplot(fig)


config_ = {
    APP_KEY: {'title': 'Data Visualizer'},
    RENDERING_KEY: {
        'plot_umap': {
            'execution': {
                'inputs': {'Xy': {ELEMENT_KEY: SelectBox, 'options': mall['Xy']},},
                'output': {ELEMENT_KEY: UmapPlotter},
            },
        },
    },
}
# ============ END FRONTEND ============

if __name__ == '__main__':
    app = mk_app([load_dataset, plot_umap], config=config_)
    app()



from typing import Iterable
from streamlitfront.elements import TextInput, SelectBox, ExecSection, TextOutput
from dataclasses import dataclass
from streamlitfront import mk_app, binder as b
from front.elements import InputBase
from i2 import Sig
from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY
import streamlit as st


def add_nums(nums_to_add: Iterable[float] = 12):
    # return sum([int(nums_to_add)])
    return nums_to_add


if not b.mall():
    b.mall = {
        'result': {},
    }

mall = b.mall()


@dataclass
class ArgsInput(InputBase):
    # func_sig: Sig = None

    def __post_init__(self):
        super().__post_init__()
        self.get_args = []
        self.inputs = []

    def render(self):
        exec_section = ExecSection(
            obj=self.get_args,
            inputs=self.inputs,
            output={ELEMENT_KEY: TextOutput},
            auto_submit=True,
            on_submit=self.populate_input,
            use_expander=False,
        )
        exec_section()
        st.write(f'inputs={self.inputs}, args={self.output}')
        return self.value()

    def populate_input(self, output):
        st.write(f'inputs={self.inputs}, args={self.output}')
        self.value.set(output)
        # self.inputs.append(item)


if not b.list_args():
    b.list_args = []


def populate_list():
    value = b.result()
    b.list_args().append(value)


if __name__ == '__main__':
    app = mk_app(
        [add_nums],
        config={
            APP_KEY: {'title': 'Rendering map'},
            RENDERING_KEY: {
                'add_nums': {
                    'execution': {
                        'inputs': {
                            'nums_to_add': {
                                ELEMENT_KEY: TextInput,
                                'value': b.result,
                                'on_value_change': populate_list
                                # "func_sig": Sig(add_nums),
                            },
                        },
                        # "on_submit": lambda x: print("done!"),
                    }
                },
            },
        },
    )
    app()
    st.write(mall)



import numpy as np
from know.base import SlabsIter
from taped import LiveWf, chunk_indices
from know.audio_to_store import get_resource


if __name__ == '__main__':
    # print(f"starting")
    # live_source = LiveWf

    # live_source = get_resource(live_source, dflt_callback=LiveWf)
    # chk_size = 100_000
    # end_idx = 800_000
    # logger = print
    # # session_id: SessionId = int(time() * 1000)
    # with live_source:
    #     # range(start, stop, step)
    #     for i, (bt, tt) in enumerate(chunk_indices(chk_size=chk_size, end_idx=end_idx)):
    #         if logger:
    #             logger(f"{i=}: {bt=}, {tt=}")
    #             print(np.mean(live_source[bt:tt]))

    from know.audio_to_store import *

    wfs = demo_live_data_acquisition(
        live_source=LiveWf(
            input_device_index=None,  # if None, will try to guess the device index
            sr=44100,
            sample_width=2,
            chk_size=4096,
            stream_buffer_size_s=60,
        ),
        store=mk_session_block_wf_store(
            rootdir=None,  # will choose one for you
            # template examples: '{session}_{block}.wav' '{session}/d/{block}.pcm', '{session}/{block}', 'demo/s_{session}/blocks/{block}.wav'
            template='{session}/d/{block}.pcm',  #
            pattern=r'\d+',
            value_trans=int,
        ),
        chk_size=100_000,
        end_idx=300_000,
        logger=print,
    )
    print(f'{len(wfs)=}')



from streamz import Stream


# TODO: most river classes support batches, e.g., learn_many, more efficiently


class RiverTransform(Stream):
    """Pass data through one or more River transforms"""

    def __init__(self, model, **kwargs):
        super().__init__(**kwargs)
        self.model = model

    def update(self, x, who=None, metadata=None):
        out = self.model.transform_one(*x)
        self.emit(out)


class RiverTrain(Stream):
    def __init__(self, model, metric=None, pass_model=False, **kwargs):
        """
        If metric and pass_model are both defaults, this is effectively
        a sink.
        :param model: river model or pipeline
        :param metric: river metric
            If given, it is emitted on every sample
        :param pass_model: bool
            If True, the (updated) model if emitted for each sample
        """
        super().__init__(**kwargs)
        self.model = model
        if pass_model and metric is not None:
            raise TypeError
        self.pass_model = pass_model
        self.metric = metric

    def update(self, x, who=None, metadata=None):
        """
        :param x: tuple
            (x, [y[, w]) floats for single sample. Include
        """
        self.model.learn_one(*x)
        if self.metric:
            yp = self.model.predict_one(x[0])
            weights = x[2] if len(x) > 1 else 1.0
            return self._emit(
                self.metric.update(x[1], yp, weights).get(), metadata=metadata
            )
        if self.pass_model:
            return self._emit(self.model, metadata=metadata)


class RiverPredict(Stream):
    def __init__(self, model, **kwargs):
        super().__init__(**kwargs)
        self.model = model

    def update(self, x, who=None, metadata=None):
        out = self.model.predict_one(x)
        return self._emit(out, metadata=metadata)



# streamlit run pydantic_streamlit_input.py

import streamlit as st
from pydantic import BaseModel
from typing import List, Dict
import streamlit_pydantic as sp
from pydantic import BaseModel, Field, HttpUrl
from enum import Enum


class SelectionValue(str, Enum):
    FOO = 'foo'
    BAR = 'bar'


class NestedData(BaseModel):
    text: str
    integerchoice: List[int]


class ExampleModel(BaseModel):
    some_text: str
    some_number: int
    some_boolean: bool
    some_list: List[int]
    some_dict: Dict[str, int]
    integer_in_range: int = Field(
        20,
        ge=10,
        lt=30,
        multiple_of=2,
        description='Number property with a limited range.',
    )
    single_selection: SelectionValue = Field(
        ..., description='Only select a single item from a set.'
    )
    nested_choice: NestedData = Field(
        ..., description='Another object embedded into this model.',
    )


st.title('Use of pydantic for complex inputs')
data = sp.pydantic_input(key='my_form', model=ExampleModel)
if data:
    st.write(data)



import pytest
from typing import List, Dict, Callable, Iterable, Union, Tuple
from plunk.sb.type_compatibility.compatible import has_compatible_type, builtins


def test_builtins_leaves():

    assert has_compatible_type(int, float)
    assert has_compatible_type(int, int)
    assert has_compatible_type(List[int], List[float])


def test_unions():
    t1 = Union[int, float]
    t2 = Union[str, bool]
    t3 = Union[float, str]
    assert has_compatible_type(t1, t3)
    assert not has_compatible_type(t1, t2)


def test_list():
    assert has_compatible_type(List[int], List[float])
    assert not has_compatible_type(List[float], List[int])


def test_dict():
    assert has_compatible_type(Dict[str, List[int]], Dict[str, List[float]])


def test_classes():
    class A:
        pass

    class B(A):
        pass

    assert has_compatible_type(A, B)


def test_tuple():
    t1 = Tuple[int, float]
    t2 = Tuple[float, float]
    assert has_compatible_type(t1, t2)


def test_callable():
    assert has_compatible_type(
        Callable[[float, float], int], Callable[[int, float], int]
    )
    # assert not has_compatible_type(List[float], List[int])



# streamlit run pydantic_streamlit_front.py

import streamlit as st
import streamlit_pydantic as sp
from streamlitfront.base import dispatch_funcs, BasePageFunc
from opyratorfront.py2pydantic import func_to_pyd_input_model_cls
from pydantic import BaseModel, Field
from i2 import name_of_obj


def foo(word: str, multiplier: int) -> str:
    return word * multiplier


class SimplePageFuncPydanticWrite(BasePageFunc):
    def __call__(self, state):
        self.prepare_view(state)
        mymodel = func_to_pyd_input_model_cls(self.func)
        name = name_of_obj(self.func)

        # data = sp.pydantic_form(key=f"my_form_{name}", model=mymodel)
        data = sp.pydantic_input(key=f'my_form_{name}', model=mymodel)

        if data:
            st.write(self.func(**data))
            # st.write(self.func(**dict(data)))
            # st.json(data.json())


configs = {'page_factory': SimplePageFuncPydanticWrite}
app = dispatch_funcs([foo], configs=configs)
app()



import threading

import cv2
import streamlit as st
from matplotlib import pyplot as plt

from streamlit_webrtc import webrtc_streamer

lock = threading.Lock()
img_container = {'img': None}


def video_frame_callback(frame):
    img = frame.to_ndarray(format='bgr24')
    with lock:
        img_container['img'] = img

    return frame


ctx = webrtc_streamer(key='example', video_frame_callback=video_frame_callback)

fig_place = st.empty()
fig, ax = plt.subplots(1, 1)

while ctx.state.playing:
    with lock:
        img = img_container['img']
    if img is None:
        continue
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    ax.cla()
    ax.hist(gray.ravel(), 256, [0, 256])
    fig_place.pyplot(fig)



"""
An app that loads either a wav file from local folder or records a sound
and visualizes the resulting numpy array 
"""
from typing import Mapping
from know.boxes import *
from functools import partial
from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY
from front.crude import Crudifier

from streamlitfront import mk_app, binder as b
from streamlitfront.elements import (
    SuccessNotification,
    PipelineMaker,
)

import streamlit as st
from dataclasses import dataclass

from typing import List


@dataclass
class Pipeline:
    steps: List[str]


def get_steps_from_selected_pipeline(pipeline):
    return pipeline.steps


def mk_pipeline_maker_app_with_mall(
    mall: Mapping, *, pipelines: str = 'pipelines', pipelines_store=None,
):
    if not b.mall():
        b.mall = mall
    mall = b.mall()

    crudifier = partial(Crudifier, mall=mall)

    pipelines_store = pipelines_store or pipelines

    @crudifier(
        param_to_mall_map=dict(pipeline=pipelines_store), output_store=pipelines_store,
    )
    def modify_pipeline(pipeline, steps):
        pipe = Pipeline(steps=steps)
        return pipe

    def view_state():
        st.write(mall)

    def on_select_pipeline(pipeline):
        b.steps_of_selected_pipeline.set(mall['pipelines'][pipeline].steps)
        st.write(f'{pipeline=} selected')

    config = {
        APP_KEY: {'title': 'Data Preparation'},
        RENDERING_KEY: {
            'modify_pipeline': {
                NAME_KEY: 'Pipeline Modify',
                'execution': {
                    'inputs': {
                        'pipeline': {
                            'value': b.selected_pipeline,
                            'on_value_change': on_select_pipeline,
                        },
                        'steps': {
                            ELEMENT_KEY: PipelineMaker,
                            'items': ['a', 'b'],
                            'steps': b.steps_of_selected_pipeline(),
                        },
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'The pipeline has been modified successfully.',
                    },
                },
            },
        },
    }

    funcs = [
        modify_pipeline,
        view_state,
    ]
    app = mk_app(funcs, config=config)

    return app


mall = dict(
    pipelines={
        'pipe1': Pipeline(steps=['a', 'b']),
        'pipe2': Pipeline(steps=['b', 'a']),
    },
)


if __name__ == '__main__':

    app = mk_pipeline_maker_app_with_mall(mall, pipelines='pipelines')

    app()



"""Components for tw_app4_drill"""

from functools import partial
from typing import Callable, Iterable, List
from io import BytesIO

from lined import LineParametrized
import numpy as np

import soundfile as sf

# --------------------------------------------------------------------------------------
# UTILS

from plunk.sb.front_demo.user_story1.utils.tools import (
    Stroll,
    scores_to_intervals,
)
from slang import fixed_step_chunker

DFLT_FEATURIZER = lambda chk: np.abs(np.fft.rfft(chk))
DFLT_CHK_SIZE = 2048
DFLT_CHK_STEP = 2048
WaveForm = Iterable[int]


def simple_chunker(it, chk_size: int = DFLT_CHK_SIZE):
    return fixed_step_chunker(it=it, chk_size=chk_size, chk_step=chk_size)


def clean_dict(kwargs):
    result = {k: v for k, v in kwargs.items() if v != ''}
    return result


def tagged_sound_to_array(train_audio: WaveForm, tag: str):
    sound, tag = train_audio, tag
    if not isinstance(sound, bytes):
        sound = sound.getvalue()

    arr = sf.read(BytesIO(sound), dtype='int16')[0]  # TODO: use recode
    return arr, tag


def tagged_sounds_to_single_array(train_audio: List[WaveForm], tag: str):
    sounds, tag = train_audio, tag
    result = []
    for sound in sounds:
        # if not isinstance(sound, bytes):
        sound = sound.getvalue()
        arr = sf.read(BytesIO(sound), dtype='int16')[0]
        result.append(arr)
    # print(np.hstack(result))
    return np.hstack(result).reshape(-1, 1), tag


def assert_dims(wfs):
    if wfs.ndim >= 2:
        wfs = wfs[:, 0]
    return wfs


featurizer = DFLT_FEATURIZER
chunker = simple_chunker
WaveForm = Iterable[int]

# --------------------------------------------------------------------------------------
# COMPONENTS

# @crudifier(
#         param_to_mall_map=dict(step_factory=step_factories), output_store=steps_store
#     )
def mk_step(step_factory: Callable, kwargs: dict):
    kwargs = clean_dict(kwargs)  # TODO improve that logic
    step = partial(step_factory, **kwargs)()

    return step


#
# @crudifier(output_store=pipelines_store,)
def mk_pipeline(steps: Iterable[Callable]):
    return LineParametrized(*steps)


# @crudifier(
#     param_to_mall_map=dict(pipeline=pipelines_store, tagged_data='sound_output'),
#     output_store='exec_outputs',
# )
def exec_pipeline(pipeline: Callable, tagged_data):

    sound, _ = tagged_sound_to_array(*tagged_data)
    wfs = np.array(sound)
    wfs = assert_dims(wfs)

    result = pipeline(wfs)
    return result


# @crudifier(
#     param_to_mall_map=dict(
#         tagged_data='sound_output', preprocess_pipeline='pipelines'
#     ),
#     output_store='learned_models',
# )
def learn_outlier_model(tagged_data, preprocess_pipeline, n_centroids=5):
    sound, tag = tagged_sounds_to_single_array(*tagged_data)
    wfs = np.array(sound)

    wfs = assert_dims(wfs)

    fvs = preprocess_pipeline(wfs)
    model = Stroll(n_centroids=n_centroids)
    model.fit(X=fvs)

    return model


# @crudifier(
#     param_to_mall_map=dict(
#         tagged_data='sound_output',
#         preprocess_pipeline='pipelines',
#         fitted_model='learned_models',
#     ),
#     output_store='models_scores',
# )
def apply_fitted_model(tagged_data, preprocess_pipeline, fitted_model):
    sound, tag = tagged_sounds_to_single_array(*tagged_data)
    wfs = np.array(sound)
    wfs = assert_dims(wfs)

    fvs = preprocess_pipeline(wfs)
    scores = fitted_model.score_samples(X=fvs)
    return scores


# @crudifier(param_to_mall_map=dict(pipeline=pipelines_store),)
def visualize_pipeline(pipeline: LineParametrized):

    return pipeline


# @crudifier(param_to_mall_map=dict(scores='models_scores'),)
def visualize_scores(scores, threshold=80, num_segs=3):

    intervals = scores_to_intervals(scores, threshold, num_segs)

    return scores, intervals


# @crudifier(output_store='sound_output')
def upload_sound(train_audio: List[WaveForm], tag: str):
    # sound, tag = train_audio, tag
    # if not isinstance(sound, bytes):
    #     sound = sound.getvalue()

    # arr = sf.read(BytesIO(sound), dtype='int16')[0]
    # return arr, tag
    return train_audio, tag


# @crudifier(param_to_mall_map={'result': 'sound_output'})
def display_tag_sound(result):
    return result



"""
An app that loads either a wav file from local folder or records a sound
and visualizes the resulting numpy array 
"""
from dataclasses import dataclass
from functools import partial
from typing import Callable, Iterable, List, Mapping
from i2 import FuncFactory, Sig
from lined import LineParametrized
import numpy as np
import matplotlib.pyplot as plt
import soundfile as sf
from io import BytesIO
from more_itertools import consecutive_groups

from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY
from front.crude import Crudifier
from front.elements import OutputBase
from streamlitfront import mk_app, binder as b
from streamlitfront.elements import (
    SelectBox,
    SuccessNotification,
    PipelineMaker,
)
from streamlitfront.elements import FileUploader
import streamlit as st

from omodel.gen_utils.chunker import fixed_step_chunker
from omodel.outliers.pystroll import OutlierModel as Stroll


DFLT_CHK_SIZE = 2048
DFLT_FEATURIZER = lambda chk: np.abs(np.fft.rfft(chk))
DFLT_NUM_OUTLIERS = 3
WaveForm = Iterable[int]


def simple_chunker(wfs, chk_size: int = DFLT_CHK_SIZE):
    return list(fixed_step_chunker(wfs, chk_size=chk_size))


def apply_func_to_index_groups(func, arr, idx_groups):
    """
    Applies func to the slices of array arr specified by the groups
    """
    result = [func(arr[gp]) for gp in idx_groups]

    return result


def arg_top_max(arr, num_elements):
    """
    returns the largest num_elements from the array arr
    """
    result = [arr.index(i) for i in sorted(arr, reverse=True)][:num_elements]
    return result


def clean_dict(kwargs):
    result = {k: v for k, v in kwargs.items() if v != ''}
    return result


def indices_for_percentile(arr, low=0, high=100):
    return np.percentile(arr, [low, high])


def consecutive_indices(arr):
    return list(map(list, consecutive_groups(arr)))


def get_groups_extremities_all(
    arr, groups_indices, func=np.mean, num_outliers=DFLT_NUM_OUTLIERS,
):

    means = apply_func_to_index_groups(func, arr, groups_indices)
    arg = arg_top_max(means, num_outliers)
    result = [
        (groups_indices[idx][0], groups_indices[idx][-1])
        for idx, _ in enumerate(means)
        if idx in arg
    ]

    return result


def scores_to_intervals(scores, high_percentile=90, num_selected=3):
    [low, high] = list(indices_for_percentile(scores, low=0, high=high_percentile))
    arr_selected = np.nonzero(scores >= high)[0]
    groups_indices = consecutive_indices(arr_selected)
    intervals_all = get_groups_extremities_all(
        scores, groups_indices, func=np.mean, num_outliers=num_selected,
    )
    result = [(a, b) for a, b in intervals_all if b > a]

    return result


def simple_featurizer(chks):
    fvs = np.array(list(map(DFLT_FEATURIZER, chks)))
    return fvs


def tagged_sound_to_array(train_audio: WaveForm, tag: str):
    sound, tag = train_audio, tag
    if not isinstance(sound, bytes):
        sound = sound.getvalue()

    arr = sf.read(BytesIO(sound), dtype='int16')[0]
    return arr, tag


def tagged_sounds_to_single_array(train_audio: List[WaveForm], tag: str):
    sounds, tag = train_audio, tag
    result = []
    for sound in sounds:
        # if not isinstance(sound, bytes):
        sound = sound.getvalue()
        arr = sf.read(BytesIO(sound), dtype='int16')[0]
        result.append(arr)
    # print(np.hstack(result))
    return np.hstack(result).reshape(-1, 1), tag


def assert_dims(wfs):
    if wfs.ndim >= 2:
        wfs = wfs[:, 0]
    return wfs


@dataclass
class AudioArrayDisplay(OutputBase):
    def render(self):
        sound, tag = self.output
        # if not isinstance(sound, str):
        if not isinstance(sound, bytes):

            sound = sound.getvalue()

        arr = sf.read(BytesIO(sound), dtype='int16')[0]
        # st.write(type(arr))
        tab1, tab2 = st.tabs(['Audio Player', 'Waveform'])
        with tab1:
            # if not isinstance(sound, bytes):
            #     sound = sound.getvalue()
            # arr = sf.read(BytesIO(sound), dtype="int16")[0]
            st.write(f'type of data={type(sound)}')

            st.audio(sound)
        with tab2:
            fig, ax = plt.subplots(figsize=(15, 5))
            ax.plot(arr, label=f'Tag={tag}')
            ax.legend()
            st.pyplot(fig)
            # st.write(arr[:10])


@dataclass
class GraphOutput(OutputBase):
    use_container_width: bool = False

    def render(self):
        # with st.expander(self.name, True): #cannot nest expanders
        dag = self.output
        st.graphviz_chart(
            figure_or_dot=dag.dot_digraph(),
            use_container_width=self.use_container_width,
        )


@dataclass
class ArrayPlotter(OutputBase):
    def render(self):
        X = self.output
        # st.write(f"Average score of session= {np.mean(X)}")
        fig, ax = plt.subplots(figsize=(15, 5))
        ax.plot(X)
        ax.legend()
        st.pyplot(fig)


@dataclass
class ArrayWithIntervalsPlotter(OutputBase):
    def render(self):
        X, intervals = self.output
        fig = pyplot_with_intervals(X, intervals)
        st.pyplot(fig)


def pyplot_with_intervals(X, intervals=None):
    min_x = np.mean(X)
    xs = list(range(len(X)))
    ys = X
    fig, ax = plt.subplots(figsize=(7, 2))
    ax.plot(xs, ys, linewidth=1)
    if intervals:
        for i, interval in enumerate(intervals):
            start, end = interval
            plt.axvspan(start, end, facecolor='g', alpha=0.5)

            ax.annotate(f'{i}', xy=(start, min_x), ha='left', va='top')
    return fig


def mk_pipeline_maker_app_with_mall(
    mall: Mapping,
    *,
    step_factories: str = 'step_factories',
    steps: str = 'steps',
    steps_store=None,
    pipelines: str = 'pipelines',
    pipelines_store=None,
    data: str = 'data',
    data_store=None,
    learned_models=None,
    models_scores=None,
):
    if not b.mall():
        b.mall = mall
    mall = b.mall()
    if not b.selected_step_factory():
        b.selected_step_factory = 'chunker'  # TODO make this dynamic

    crudifier = partial(Crudifier, mall=mall)

    steps_store = steps_store or steps
    data_store = data_store or data
    pipelines_store = pipelines_store or pipelines

    @crudifier(
        param_to_mall_map=dict(step_factory=step_factories), output_store=steps_store
    )
    def mk_step(step_factory: Callable, kwargs: dict):
        kwargs = clean_dict(kwargs)  # TODO improve that logic
        step = partial(step_factory, **kwargs)()

        return step

    #
    @crudifier(output_store=pipelines_store,)
    def mk_pipeline(steps: Iterable[Callable]):
        return LineParametrized(*steps)

    @crudifier(
        param_to_mall_map=dict(pipeline=pipelines_store, tagged_data='sound_output'),
        output_store='exec_outputs',
    )
    def exec_pipeline(pipeline: Callable, tagged_data):

        sound, _ = tagged_sound_to_array(*tagged_data)
        wfs = np.array(sound)
        wfs = assert_dims(wfs)

        result = pipeline(wfs)
        return result

    @crudifier(
        param_to_mall_map=dict(
            tagged_data='sound_output', preprocess_pipeline='pipelines'
        ),
        output_store='learned_models',
    )
    def learn_outlier_model(tagged_data, preprocess_pipeline, n_centroids=5):
        sound, tag = tagged_sounds_to_single_array(*tagged_data)
        wfs = np.array(sound)

        wfs = assert_dims(wfs)

        fvs = preprocess_pipeline(wfs)
        model = Stroll(n_centroids=n_centroids)
        model.fit(X=fvs)

        return model

    @crudifier(
        param_to_mall_map=dict(
            tagged_data='sound_output',
            preprocess_pipeline='pipelines',
            fitted_model='learned_models',
        ),
        output_store='models_scores',
    )
    def apply_fitted_model(tagged_data, preprocess_pipeline, fitted_model):
        sound, tag = tagged_sounds_to_single_array(*tagged_data)
        wfs = np.array(sound)
        wfs = assert_dims(wfs)

        fvs = preprocess_pipeline(wfs)
        scores = fitted_model.score_samples(X=fvs)
        return scores

    @crudifier(param_to_mall_map=dict(pipeline=pipelines_store),)
    def visualize_pipeline(pipeline: LineParametrized):

        return pipeline

    @crudifier(param_to_mall_map=dict(scores='models_scores'),)
    def visualize_scores(scores, threshold=80, num_segs=3):

        intervals = scores_to_intervals(scores, threshold, num_segs)

        return scores, intervals

    @crudifier(output_store='sound_output')
    def upload_sound(train_audio: List[WaveForm], tag: str):
        # sound, tag = train_audio, tag
        # if not isinstance(sound, bytes):
        #     sound = sound.getvalue()

        # arr = sf.read(BytesIO(sound), dtype='int16')[0]
        # return arr, tag
        return train_audio, tag

    @crudifier(param_to_mall_map={'result': 'sound_output'})
    def display_tag_sound(result):
        return result

    def get_step_name(step):
        return [k for k, v in mall[steps].items() if v == step][0]

    def get_selected_step_factory_sig():
        selected_step_factory = mall['step_factories'].get(
            b.selected_step_factory.get()
        )
        if selected_step_factory:
            return Sig(selected_step_factory)

    config = {
        APP_KEY: {'title': 'Data Preparation'},
        RENDERING_KEY: {
            'upload_sound': {
                # NAME_KEY: "Data Loader",
                'execution': {
                    'inputs': {
                        'train_audio': {
                            ELEMENT_KEY: FileUploader,
                            'type': 'wav',
                            'accept_multiple_files': True,
                        },
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'Wav files loaded successfully.',
                    },
                },
            },
            'display_tag_sound': {
                'execution': {'output': {ELEMENT_KEY: AudioArrayDisplay,},},
            },
            'mk_step': {
                NAME_KEY: 'Pipeline Step Maker',
                'execution': {
                    'inputs': {
                        'step_factory': {'value': b.selected_step_factory,},
                        'kwargs': {'func_sig': get_selected_step_factory_sig},
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'The step has been created successfully.',
                    },
                },
            },
            'mk_pipeline': {
                NAME_KEY: 'Pipeline Maker',
                'execution': {
                    'inputs': {
                        steps: {
                            ELEMENT_KEY: PipelineMaker,
                            'items': list(mall[steps].values()),
                            'serializer': get_step_name,
                        },
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'The pipeline has been created successfully.',
                    },
                },
            },
            'exec_pipeline': {
                NAME_KEY: 'Pipeline Executor',
                'execution': {
                    'inputs': {
                        'pipeline': {'value': b.selected_pipeline,},
                        'data': {
                            ELEMENT_KEY: SelectBox,
                            'options': mall['sound_output'],
                        },
                    }
                },
            },
            'visualize_pipeline': {
                NAME_KEY: 'Pipeline Visualization',
                'execution': {
                    'inputs': {'pipeline': {'value': b.selected_pipeline,},},
                    'output': {
                        ELEMENT_KEY: GraphOutput,
                        NAME_KEY: 'Flow',
                        'use_container_width': True,
                    },
                },
            },
            'visualize_scores': {
                NAME_KEY: 'Scores Visualization',
                'execution': {'output': {ELEMENT_KEY: ArrayWithIntervalsPlotter,},},
            },
            'simple_model': {
                NAME_KEY: 'Learn model',
                'execution': {'output': {ELEMENT_KEY: ArrayPlotter,},},
            },
            'apply_fitted_model': {
                NAME_KEY: 'Apply model',
                'execution': {'output': {ELEMENT_KEY: ArrayPlotter,},},
            },
        },
    }

    funcs = [
        upload_sound,
        # display_tag_sound,
        mk_step,
        mk_pipeline,
        learn_outlier_model,
        apply_fitted_model,
        # exec_pipeline,
        visualize_pipeline,
        visualize_scores,
    ]
    app = mk_app(funcs, config=config)

    return app


mall = dict(
    # Factory Input Stores
    sound_output=dict(),
    step_factories=dict(
        # ML
        chunker=FuncFactory(simple_chunker),
        featurizer=FuncFactory(simple_featurizer),
    ),
    # Output Store
    data=dict(),
    steps=dict(),
    pipelines=dict(),
    exec_outputs=dict(),
    learned_models=dict(),
    models_scores=dict(),
)

# crudifier = partial(prepare_for_crude_dispatch, mall=mall)

# step_factories = dict(
#     # ML
#     chunker=FuncFactory(simple_chunker),
#     featurizer=FuncFactory(simple_featurizer),
# )

# mall['step_factories'] = step_factories


if __name__ == '__main__':

    app = mk_pipeline_maker_app_with_mall(
        mall, step_factories='step_factories', steps='steps', pipelines='pipelines'
    )

    app()



from dataclasses import dataclass
from typing import Iterable
from meshed import code_to_dag, DAG
from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY, OBJ_KEY
from collections.abc import Callable
from front.crude import prepare_for_crude_dispatch
from streamlitfront.elements import TextInput, SelectBox, FloatSliderInput
from front.elements import OutputBase
from streamlitfront.base import mk_app
from streamlitfront.examples.util import Graph
import base64
from streamlitfront.elements import (
    AudioRecorder,
    FileUploader,
    MultiSourceInput,
)
import streamlit as st
import matplotlib.pyplot as plt
import soundfile as sf
from io import BytesIO
from front.elements import InputBase
from streamlitfront.elements.js import mk_element_factory
import numpy as np
import os
import streamlit.components.v1 as components


# st_audiorec = mk_element_factory("st_audiorec")

parent_dir = os.path.dirname(os.path.abspath(__file__))
# Custom REACT-based component for recording client audio in browser
build_dir = os.path.join(parent_dir, 'st_audiorec/frontend/build')
# specify directory and initialize st_audiorec object functionality
st_audiorec = components.declare_component('st_audiorec', path=build_dir)
# base64data_audio = st_audiorec()

# # APPROACH: DECODE BASE64 DATA FROM return value
# st.write(base64data_audio)
# if (
#     (base64data_audio != None)
#     and (base64data_audio != "")
#     and ("test" not in base64data_audio)
# ):
#     # decoding process of base64 string to wav file
#     with st.spinner("Decoding audio data..."):
#         base64data_audio = base64data_audio.replace("data:audio/wav;base64,", "")
#         st.write(base64data_audio)  # remove metadata header of base64 string

#         audiofile_name = "temp.wav"
#         wav_file = open(audiofile_name, "wb")
#         decode_string = base64.b64decode(base64data_audio + "==")
#         wav_file.write(decode_string)

#     audiofile_path = os.path.join(os.getcwd(), audiofile_name)
#     st.write(audiofile_path)
#     st.audio(audiofile_path)

val = st_audiorec()
# web component returns arraybuffer from WAV-blob
st.write('Audio data received in the Python backend will appear below this message ...')
st.write(val['arr'])
# if not isinstance(val, dict):  # retrieve audio data

# # wav_bytes contains audio data in format to be further processed
# # display audio data as received on the Python side
# st.audio(wav_bytes, format="audio/wav")
# st.write(val)



from functools import partial
from front.crude import Crudifier

mall = dict(
    a=dict(one=1, two=2), b=dict(three=3, four=4), foo_output=dict(), bar_output=dict(),
)
crudifier = partial(Crudifier, mall=mall, include_stores_attribute=True)


@crudifier(param_to_mall_map=['a', 'b'], output_store='foo_output')
def foo(a, b: float):
    """This is foo. It computes something"""
    return a + b


@crudifier(param_to_mall_map=['foo_output'], output_store='bar_output')
def bar(foo_output: float):
    return foo_output * 10


funcs = [foo, bar]



from warnings import warn


def deprecate_attr(attr_name, attr_replacement_name):
    """Deprecates an attribute by printing a deprecation message and bind the usage of
    this attribute with another non-deprecated attribute.

    >>> @deprecate_attr('deprecated_attr', 'valid_attr')
    ... class Foo:
    ...     valid_attr = 'Some valid attribute'

    >>> foo = Foo()
    >>> foo.deprecated_attr
    'Some valid attribute'

    >>> foo.deprecated_attr = 'Got deprected attribute'
    >>> foo.deprecated_attr 
    'Got deprected attribute'

    >>> foo.valid_attr
    'Got deprected attribute'
    """

    def wrapper(cls):
        def warn_deprecated_sig_attr():
            warn(
                f'"{cls.__qualname__}.{attr_name}" is deprecated, consider using "{cls.__qualname__}.{attr_replacement_name}" instead.',
                DeprecationWarning,
                stacklevel=2,
            )

        def get_attr(self):
            warn_deprecated_sig_attr()
            return getattr(self, attr_replacement_name)

        def set_attr(self, value):
            setattr(self, attr_replacement_name, value)

        prop = property(get_attr, set_attr)
        setattr(cls, attr_name, prop)
        return cls

    return wrapper



import numpy as np
from metaflow import FlowSpec, step
from py2store import myconfigs
from graze import graze
import soundfile as sf
from dol import FilesOfZip, wrap_kvs, filt_iter
from io import BytesIO


DFLT_FEATURIZER = lambda chk: np.abs(np.fft.rfft(chk))


config_filename = 'phone_digits_block.json'
DFLT_LOCAL_SOURCE_DIR = myconfigs.get_config_value(config_filename, 'local_source_dir')


def mk_fvs(chk_tags, featurizer=DFLT_FEATURIZER):
    for chk, tag in chk_tags:
        yield featurizer(chk), tag


class ClassifierTrainFlow(FlowSpec):
    @step
    def start(self):  # A
        from odat.mdat.phone_digits_block import mk_dacc

        dacc = mk_dacc()
        self.chk_tag = dacc.chk_tag_gen()

        X, y = list(chk_tag)
        (
            self.train_data,
            self.test_data,
            self.train_labels,
            self.test_labels,
        ) = train_test_split(X, y, test_size=0.4, random_state=0)
        self.next(self.mk_fvs)

    @step
    def mk_fvs(self):  # B
        self.fvs = mk_fvs(self.chk_tag)

        fvs_train = np.array(list(map(DFLT_FEATURIZER, list_chks[:100])))
        # from sklearn.neighbors import KNeighborsClassifier

        # self.model = KNeighborsClassifier()
        # self.model.fit(self.train_data, self.train_labels)
        # self.next(self.choose_model)

    @step
    def train_test(self):
        pass

    @step
    def train_svm(self):  # B
        from sklearn import svm

        self.model = svm.SVC(kernel='poly')  # D
        self.model.fit(self.train_data, self.train_labels)
        self.next(self.choose_model)

    @step
    def choose_model(self, inputs):  # B
        def score(inp):
            return inp.model, inp.model.score(inp.test_data, inp.test_labels)

        self.results = sorted(map(score, inputs), key=lambda x: -x[1])
        self.model = self.results[0][0]
        self.next(self.end)

    @step
    def end(self):  # C
        print('Scores:')
        print('\n'.join('%s %f' % res for res in self.results))


if __name__ == '__main__':
    ClassifierTrainFlow()



import streamlit as st
import pandas as pd
import matplotlib.pyplot as plt
import soundfile as sf
from graze import graze
from dol import FilesOfZip, wrap_kvs, filt_iter
from io import BytesIO
from st_aggrid import GridOptionsBuilder, AgGrid, DataReturnMode, GridUpdateMode


DFLT_URL = 'https://www.dropbox.com/s/gnucwiq8wl7hes8/phone_digits_train.zip?dl=0'

# make some data
def my_obj_of_data(b):
    # return sf.read(BytesIO(b), dtype="float32")[0]
    return sf.read(BytesIO(b), dtype='int16')[0]


@wrap_kvs(obj_of_data=my_obj_of_data)
@filt_iter(filt=lambda x: not x.startswith('__MACOSX') and x.endswith('.wav'))
class WfStore(FilesOfZip):
    """Waveform access. Keys are .wav filenames and values are numpy arrays of int16 waveform."""

    pass


def keys_to_cols(all_keys):
    arr = []
    for key in all_keys:
        train_info, sref = key.split('/')
        arr.append((key, sref[0], train_info))
    df = pd.DataFrame(arr, columns=['key', 'tag', 'train'])
    return df


def mk_store_and_annots(dropbox_url=DFLT_URL):
    wf_store = WfStore(graze(dropbox_url))
    all_keys = list(wf_store.keys())
    df = keys_to_cols(all_keys)

    return wf_store, df


def display_selected_rows(wf_store, selected):
    for row in selected:
        key = row['key']
        tag = row['tag']
        arr = wf_store[key]
        fig, ax = plt.subplots(figsize=(15, 5))
        ax.plot(arr, label=f'Tag={tag}')
        ax.legend()
        st.pyplot(fig)


wf_store, df = mk_store_and_annots()

# The App
st.set_page_config(layout='wide')
st.title('Wav Store explorer')

gb = GridOptionsBuilder.from_dataframe(df)


def configure_grid(gb):
    selection_mode = 'multiple'
    groupSelectsChildren = True
    groupSelectsFiltered = True
    gb.configure_selection(selection_mode)
    gb.configure_selection(
        selection_mode,
        use_checkbox=True,
        groupSelectsChildren=groupSelectsChildren,
        groupSelectsFiltered=groupSelectsFiltered,
    )
    return gb


gb = configure_grid(gb)
gb.configure_grid_options(domLayout='normal')
gridOptions = gb.build()

return_mode = 'AS_INPUT'
return_mode_value = DataReturnMode.__members__[return_mode]
update_mode = 'MODEL_CHANGED'
update_mode_value = GridUpdateMode.__members__[update_mode]


col1, col2 = st.columns(2)

with col1:
    st.header('Display data')

    grid_response = AgGrid(
        df,
        gridOptions=gridOptions,
        width='100%',
        data_return_mode=return_mode_value,
        update_mode=update_mode_value,
    )

    df_r = grid_response['data']
    selected = grid_response['selected_rows']
    selected_df = pd.DataFrame(selected).apply(pd.to_numeric, errors='coerce')

with col2:
    st.header('Display selection')

    if selected:

        display_selected_rows(wf_store, selected)



"""StreamViz LiveGraph
[x] Slabs signal source and process chunks for graphs; initially volume and zero crossing ratio.
[ ] Add graph data to a buffer that can be queried by time or index.
[ ] Add to webservice.
[ ] Visualize graph.
"""
from math import ceil
from pathlib import Path
from tempfile import TemporaryDirectory
from time import sleep, time
from typing import Any, Callable, Union, Optional

import numpy as np
from audiostream2py import PyAudioSourceReader, get_input_device_index
from know.base import SlabsIter, IteratorExit
from know.util import ContextualFunc
from recode import mk_codec
from stream2py import SourceReader, StreamBuffer
from stream2py.utility.typing_hints import ComparableType

from plunk.ap.live_graph.audio_store import BulkStore, WavFileStore, N_TO_BULK
from plunk.ap.snippets import if_no_none

codec = mk_codec('h')


@if_no_none
def volume(wf):
    return np.std(wf)


@if_no_none
def zero_crossing_ratio(wf):
    return sum(np.diff(np.array(wf) > 0).astype(int)) / (len(wf) - 1)


@if_no_none
def mean_score(scores):
    return np.mean(scores)


GRAPH_TYPES = {
    'volume': {'function': volume, 'plot': 'line'},
    'zero_crossing_ratio': {'function': zero_crossing_ratio, 'plot': 'line'},
    'mean_score': {'function': mean_score, 'plot': 'line'},
}


@if_no_none
def audio_to_wf(audio):
    _, wf_bytes, *_ = audio
    wf = codec.decode(wf_bytes)
    return wf


def audio_it(
    input_device=None,
    rate=44100,
    width=2,
    channels=1,
    frames_per_buffer=44100,  # same as sample rate for 1 second intervals
    seconds_to_keep_in_stream_buffer=60,
    graph_types=(*GRAPH_TYPES,),
    audio_store: BulkStore = None,
    **additional_components,
):
    input_device_index = get_input_device_index(input_device=input_device)
    maxlen = PyAudioSourceReader.audio_buffer_size_seconds_to_maxlen(
        buffer_size_seconds=seconds_to_keep_in_stream_buffer,
        rate=rate,
        frames_per_buffer=frames_per_buffer,
    )

    audio_buffer = PyAudioSourceReader(
        rate=rate,
        width=width,
        channels=channels,
        unsigned=True,
        input_device_index=input_device_index,
        frames_per_buffer=frames_per_buffer,
    ).stream_buffer(maxlen)

    store_components = {}
    if audio_store is not None:

        @if_no_none
        def _append(timestamp, wf):
            return audio_store.append({'timestamp': timestamp, 'wf': wf})

        store_audio = ContextualFunc(_append, (audio_store,))
        store_components.update(
            audio_store_instance=lambda: audio_store, _store_audio=store_audio,
        )

    def stop_if_audio_not_running(audio_reader_instance):
        if audio_reader_instance is not None and not audio_reader_instance.is_running:
            raise IteratorExit("audio isn't running anymore!")

    @if_no_none
    def get_timestamp(audio):
        return audio[0]

    return SlabsIter(
        audio=audio_buffer,
        timestamp=get_timestamp,
        wf=audio_to_wf,
        audio_reader_instance=lambda: audio_buffer,
        _audio_stop=stop_if_audio_not_running,
        **store_components,
        **additional_components,
        **{k: v.get('function') for k, v in GRAPH_TYPES.items() if k in graph_types},
    )


class SlabsSourceReader(SourceReader):
    def __init__(
        self, slabs_it: SlabsIter, key: Union[str, Callable], post_read: Callable = None
    ):
        self.slabs_it = slabs_it
        self._key = key
        self.post_read_func = post_read

    def open(self) -> None:
        self.slabs_it.open()

    def close(self) -> None:
        self.slabs_it.close()

    def key(self, data: Any) -> ComparableType:
        if isinstance(self._key, str):
            return data.get(self._key)
        return self._key(data)

    def read(self) -> Optional[dict]:
        data = next(self.slabs_it)
        if self.post_read_func is not None:
            return self.post_read_func(data)
        return data


DATA_KEYS = (
    'timestamp',
    *GRAPH_TYPES,
)


@if_no_none
def post_read_data(data, data_keys=DATA_KEYS) -> Optional[dict]:
    """Used by stream2py to filter data

    :param data: SourceReader read data
    :param data_keys: list of keys to keep from Slabs data dict
    :return:
    """
    if data.get('timestamp') is None:
        return None
    formatted_data = {k: data.get(k) for k in data_keys}
    return formatted_data


def mk_live_graph_data_buffer(
    input_device=None,
    rate=44100,
    width=2,
    channels=1,
    frames_per_buffer=44100,
    seconds_to_keep_in_stream_buffer=60,
    graph_types=(*GRAPH_TYPES,),
    **additional_components,
) -> StreamBuffer:
    maxlen = PyAudioSourceReader.audio_buffer_size_seconds_to_maxlen(
        buffer_size_seconds=seconds_to_keep_in_stream_buffer,
        rate=rate,
        frames_per_buffer=frames_per_buffer,
    )
    return SlabsSourceReader(
        slabs_it=audio_it(
            input_device,
            rate,
            width,
            channels,
            frames_per_buffer,
            seconds_to_keep_in_stream_buffer,
            graph_types,
            **additional_components,
        ),
        key='timestamp',
        post_read=post_read_data,
    ).stream_buffer(maxlen)


def _test_audio_it(
    input_device=None,
    graph_types=(*GRAPH_TYPES,),
    test_length=22,
    **additional_components,
):
    with TemporaryDirectory() as tmpdirname:
        audio_store = WavFileStore(rootdir=tmpdirname)

        i = 0
        with audio_it(
            input_device=input_device,
            graph_types=graph_types,
            audio_store=audio_store,
            **additional_components,
        ) as a_it:
            for a in a_it:
                if a.get('audio'):
                    slab = post_read_data(a)
                    slab.update()
                    print('slab', i, slab)
                    print(f'{len(a.get("audio_store_instance")._tracks)=}')
                    assert all(k in slab for k in DATA_KEYS)
                    i += 1
                    if i >= test_length:
                        break

        n_tracks = len(audio_store._tracks)
        print(f'{n_tracks=}')
        assert n_tracks == 0, 'Tracking should flush after audio_it breaks out of loop'
        print(f'{len(audio_store)=}')
        assert len(audio_store) == ceil(test_length / N_TO_BULK), (
            'Failed to bulk items: ' f'{len(a)=} == {ceil(test_length / N_TO_BULK)=}'
        )

        if n_tracks:
            audio_store.flush()
            print(f'{len(audio_store)=}')
            print(list(audio_store))
            print(list(Path(tmpdirname).iterdir()))
            assert len(audio_store) == ceil(test_length / N_TO_BULK), (
                'Failed to bulk items: '
                f'{len(a)=} == {ceil(test_length / N_TO_BULK)=}'
            )

    assert Path(tmpdirname).is_dir() is False
    print('Test passed: _test_audio_it')


def _test_live_graph_data_buffer(
    input_device=None, graph_types=(*GRAPH_TYPES,), test_length=22
):
    recording_devices = PyAudioSourceReader.list_recording_devices()
    recording_devices.append(None)
    print(recording_devices)
    assert input_device in recording_devices, 'Selected input device not found.'

    start = time() * 1e6
    with mk_live_graph_data_buffer(
        input_device=input_device, graph_types=graph_types
    ) as slab_buffer:
        i = 0
        slab_reader = slab_buffer.mk_reader()
        while slab_buffer.is_running:
            if (slab := slab_reader.read(ignore_no_item_found=True)) is not None:
                print('slab', i, slab)
                assert all(k in slab for k in DATA_KEYS)
                i += 1
            else:
                sleep(0.1)

            if i == test_length:
                break

        stop = time() * 1e6
        print(f'{(len(slab_reader.range(start, stop)) == i)=}')
        assert len(slab_reader.range(start, stop)) == i
    print('Test passed: _test_live_graph_data_buffer')


if __name__ == '__main__':
    input_device = 'NexiGo N930AF FHD Webcam Audio'
    graph_types = 'volume'

    _test_audio_it(input_device=input_device, graph_types=graph_types)
    _test_live_graph_data_buffer(input_device=input_device, graph_types=graph_types)



from dataclasses import dataclass
from functools import cached_property
from typing import (
    Any,
    Mapping,
    Union,
    Hashable,
    List,
    TypeVar,
    Protocol,
    Tuple,
    Callable,
)
import operator

import streamlit as st
from boltons.typeutils import make_sentinel
from front.elements import OutputBase, InputBase
from functools import reduce


class _RenderInput(Protocol):
    def __call__(_self, self: 'StoreExplorerInput', depth: 'Depth', obj: Any) -> None:
        ...


class _RenderOutput(Protocol):
    def __call__(_self, self: 'StoreExplorerOutput', obj: Any) -> None:
        ...


RenderInput = TypeVar('RenderInput', bound=_RenderInput)
RenderOutput = TypeVar('RenderOutput', bound=_RenderOutput)
Depth = TypeVar('Depth', bound=int)
DepthKey = TypeVar('DepthKey', bound=Union[Hashable, int])
NOT_SELECTED = make_sentinel('Not Selected', 'Not Selected')
STORE_EXPLORER_STATE = '__STORE_EXPLORER_STATE__'


def get_store_explorer_state(defaults: dict = None):
    from streamlitfront import binder as b

    if not b.store_explorer_state():
        b.store_explorer_state = defaults
    m = b.store_explorer_state()
    try:
        if isinstance(defaults, dict) and not all(k in m for k in defaults):
            m.update(defaults)
    except Exception as e:
        raise ValueError(f'get_store_explorer_state({defaults=})') from e
    return m


_state = get_store_explorer_state({'depth_keys': []})


def get_by_path(root, items):
    """Access a nested object in root by item sequence."""
    return reduce(operator.getitem, items, root)


def del_by_path(root, items):
    """Delete a key-value in a nested object in root by item sequence."""
    del get_by_path(root, items[:-1])[items[-1]]


@dataclass
class StoreExplorerInput(InputBase):
    mall: Mapping = dict

    def __post_init__(self):
        super().__post_init__()
        print(f'0-{self.mall=}')
        if isinstance(self.mall, Callable):
            self.mall = self.mall()
        print(f'1-{self.mall=}')

    @cached_property
    def stores(self) -> Mapping:
        """Remove STORE_EXPLORER_STATE from selection"""
        return {k: v for k, v in self.mall.items() if k is not STORE_EXPLORER_STATE}

    @property
    def depth_keys(self) -> List[DepthKey]:
        return _state['depth_keys']

    @depth_keys.setter
    def depth_keys(self, value: List[DepthKey]):
        _state['depth_keys'] = value

    @depth_keys.deleter
    def depth_keys(self):
        del _state['depth_keys']

    @staticmethod
    def st_key(depth) -> str:
        return f'depth {depth}'

    @staticmethod
    def get_item(obj: Mapping, key: Union[Hashable, int]) -> Any:
        if key is NOT_SELECTED:
            return NOT_SELECTED
        try:
            return obj[key]
        except KeyError as e:
            raise KeyError(f'No {key=} in {obj=}') from e

    def selectbox(self, depth: Depth, options: List[DepthKey]):
        st.selectbox(
            key := self.st_key(depth),
            options=options,
            key=key,
            on_change=self._update_depth_keys,
            args=(depth,),
        )

    def _update_depth_keys(self, depth: Depth) -> None:
        """Update depth keys on change in streamlit component

        :param depth: number of steps into the store
        :return: None
        """
        self.depth_keys = self.depth_keys[:depth]
        if (dk := st.session_state.get(self.st_key(depth))) is not NOT_SELECTED:
            self.depth_keys.append(dk)

    def render(self):
        obj = self.stores
        depth = -1
        for depth, k in enumerate(self.depth_keys):
            action: RenderInput = getattr(self, type(obj).__name__, self.default)
            action(depth, obj)
            obj = self.get_item(obj, k)
        depth += 1
        action: RenderInput = getattr(self, type(obj).__name__, self.default)
        action(depth, obj)
        before_deletion = self.depth_keys

        def rm():
            stuff = del_by_path(self.mall, self.depth_keys)
            self.depth_keys = self.depth_keys[:-1]
            st.write(f'{stuff=}')

        st.button('delete key', on_click=rm)
        return before_deletion

    def dict(self, depth: Depth, obj: dict):
        """RenderInput for dict"""
        if len(obj) != 0:
            self.selectbox(depth, [NOT_SELECTED, *obj])

    def list(self, depth: Depth, obj: list):
        """RenderInput for list"""
        if len(obj) != 0:
            self.selectbox(depth, options=[NOT_SELECTED, *(i for i in range(len(obj)))])

    def default(self, depth: Depth, obj: Any):
        """Default RenderInput"""


@dataclass
class StoreExplorerOutput(OutputBase):
    output: Tuple[List[DepthKey], Any] = ((), {})
    write_depth_keys: Union[bool, Callable[[List[DepthKey]], None]] = True

    def __post_init__(self):
        if self.write_depth_keys is True:
            self.write_depth_keys = self.default_write_depth_keys
        elif self.write_depth_keys is False:
            self.write_depth_keys = lambda depth_keys: None

    @property
    def depth_keys(self) -> List[DepthKey]:
        return self.output[0]

    @property
    def node(self) -> Any:
        return self.output[1]

    def render(self):
        self.write_depth_keys(self.depth_keys)
        _render: RenderOutput = getattr(self, type(self.node).__name__, self.default)
        _render(self.node)

    @staticmethod
    def default_write_depth_keys(depth_keys):
        st.write(f'#### Store[{"][".join(str(k) for k in depth_keys)}] Value')

    def dict(self, obj: dict):
        """Render for dict"""
        if STORE_EXPLORER_STATE in obj:
            obj = {k: v for k, v in obj.items() if k != STORE_EXPLORER_STATE}

        self.default(obj)

    def list(self, obj: list):
        """Render for list"""
        self.default(obj)

    def default(self, obj: Any):
        """Default RenderInput"""
        st.write(obj)



from meshed import FuncNode, DAG
import streamlit as st
from plunk.sb.front_experiments.streamlitfront_dataprep.data_prep2 import (
    data_from_wav_folder,
    data_from_csv,
    store_to_key_fvs,
    key_fvs_to_tag_fvs,
    mk_Xy,
)
from odat.mdat.vacuum import (
    DFLT_ANNOTS_COLS,
    DFLT_CHUNKER,
    DFLT_FEATURIZER,
)
from streamlitfront.examples.util import Graph
from streamlitfront.elements import TextInput, SelectBox, FloatSliderInput

from front.spec_maker_base import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY
from streamlitfront.base import mk_app

DFLT_CHUNKER_MAKER = lambda: DFLT_CHUNKER
DFLT_FEATURIZER_MAKER = lambda: DFLT_FEATURIZER

FixedSizeChunker = DFLT_CHUNKER
Featurizer = DFLT_FEATURIZER

if 'mall' not in st.session_state:
    st.session_state['mall'] = dict(
        # train_audio={},
        # tag={},
        # unused_store={"to": "illustrate"},
        global_input={}
    )

mall = st.session_state['mall']
mall['global_input'] = ['a name', 3]


def funcnode_maker(name):
    kwargs = metadata[name]
    return FuncNode(**kwargs)


def dag_maker(funcnames_list):
    func_nodes = list(map(funcnode_maker, funcnames_list))
    return DAG(func_nodes)


def f(msg):
    return msg + 'bob'


def g(x):
    return int(x) + 1


def mul(msg_out, multiplier):
    return msg_out * multiplier


metadata = {
    'f': {'func': f, 'out': 'msg_out'},
    'g': {'func': g, 'out': 'multiplier'},
    'mul': {'func': mul, 'out': 'result'},
}

# Make a dag from only typing info
# my_chunker: Chunker    -->
funcnames_list = ['f', 'g', 'mul']


def delegate_input(input_nodes):
    dflt_val = {
        ELEMENT_KEY: SelectBox,
        'options': mall['global_input'],
    }

    return {k: dflt_val for k in input_nodes}


if __name__ == '__main__':

    dag = dag_maker(funcnames_list)
    var_nodes = dag.var_nodes
    print(var_nodes)
    nodes_list = ['x', 'msg']
    print(dag.synopsis_string())

    config_ = {
        APP_KEY: {'title': 'Simple Load and Display'},
        RENDERING_KEY: {
            DAG: {
                'graph': {ELEMENT_KEY: Graph, NAME_KEY: 'Flow',},
                'execution': {'inputs': delegate_input(nodes_list),},
            },
        },
    }

    app = mk_app([dag], config=config_)
    app()



from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY

from streamlitfront.elements import FloatSliderInput, FloatInput
from streamlitfront.base import mk_app
from front.crude import prepare_for_crude_dispatch
import streamlit as st


if 'mall' not in st.session_state:
    st.session_state['mall'] = {'train_test_proportion': {'train_test_proportion': 0.8}}

mall = st.session_state['mall']


def foo(a: int = 1, b: int = 2, c=3):
    """This is foo. It computes something"""
    return (a * b) + c


def bar(x, greeting):
    """bar greets its input"""
    return f'{greeting} {x}'


def set_train_test_proportion(train_proportion):
    mall['train_test_proportion'] = train_proportion
    return train_proportion


app = mk_app(
    [foo, bar, set_train_test_proportion],
    config={
        APP_KEY: {'title': 'My app'},
        RENDERING_KEY: {
            'set_train_test_proportion': {
                'execution': {
                    'inputs': {'train_proportion': {ELEMENT_KEY: FloatSliderInput,},}
                },
            },
        },
    },
)
app()
st.write(mall)



from dataclasses import dataclass
from front.elements import OutputBase
import streamlit as st
import matplotlib.pyplot as plt
import numpy as np
from sklearn.preprocessing import normalize
import soundfile as sf
from io import BytesIO


@dataclass
class AudioArrayDisplay(OutputBase):
    def render(self):
        sound, tag = self.output
        # if not isinstance(sound, str):
        if not isinstance(sound, bytes):

            sound = sound.getvalue()

        arr = sf.read(BytesIO(sound), dtype='int16')[0]
        # st.write(type(arr))
        tab1, tab2 = st.tabs(['Audio Player', 'Waveform'])
        with tab1:
            # if not isinstance(sound, bytes):
            #     sound = sound.getvalue()
            # arr = sf.read(BytesIO(sound), dtype="int16")[0]
            st.write(f'type of data={type(sound)}')

            st.audio(sound)
        with tab2:
            fig, ax = plt.subplots(figsize=(15, 5))
            ax.plot(arr, label=f'Tag={tag}')
            ax.legend()
            st.pyplot(fig)
            # st.write(arr[:10])


@dataclass
class GraphOutput(OutputBase):
    use_container_width: bool = False

    def render(self):
        # with st.expander(self.name, True): #cannot nest expanders
        dag = self.output
        st.graphviz_chart(
            figure_or_dot=dag.dot_digraph(),
            use_container_width=self.use_container_width,
        )


@dataclass
class ArrayPlotter(OutputBase):
    def render(self):
        X = self.output
        # st.write(f"Average score of session= {np.mean(X)}")
        fig, ax = plt.subplots(figsize=(15, 5))
        ax.plot(X)
        ax.legend()
        st.pyplot(fig)


@dataclass
class ArrayWithIntervalsPlotter(OutputBase):
    def render(self):
        X, intervals = self.output
        fig = pyplot_with_intervals(X, intervals)
        st.pyplot(fig)


def pyplot_with_intervals(X, intervals=None):
    min_x = np.mean(X)
    xs = list(range(len(X)))
    ys = X
    fig, ax = plt.subplots(figsize=(7, 2))
    ax.plot(xs, ys, linewidth=1)
    if intervals:
        for i, interval in enumerate(intervals):
            start, end = interval
            plt.axvspan(start, end, facecolor='g', alpha=0.5)

            ax.annotate(f'{i}', xy=(start, min_x), ha='left', va='top')
    return fig



from functools import partial
import streamlit as st

from front.crude import prepare_for_crude_dispatch

# ============ BACKEND ============

import numpy as np


def apply_func_to_wf(wf, func):
    print(f'{wf=}, {func=}')
    return func(wf)


mall = dict(func=dict(volume=np.std, bull=np.mean,),)
# crudify = Crudifier(mall)

import front as f
import streamlitfront.elements as sf

from i2 import wrap, Pipe
from recode import decode_wav_bytes


class MyFileUploader(sf.FileUploader):
    def render(self):
        print('... in render with {self=} and {uploaded_file=}')
        uploaded_file = super().render()
        if uploaded_file is not None:
            b = decode_wav_bytes(uploaded_file.read())
            print(f'{len(b)=}, {type(b)=}')
            return decode_wav_bytes(uploaded_file.read())


from front.crude import prepare_for_crude_dispatch

apply_func_to_wf = prepare_for_crude_dispatch(
    apply_func_to_wf, param_to_mall_map={'func': 'func'}, mall=mall
)

# my_file_uploader = wrap(sf.FileUploader, egress=)

config_ = {
    # f.APP_KEY: {'title': 'apply_func_to_wf'},
    f.RENDERING_KEY: {
        'apply_func_to_wf': {
            'execution': {
                'inputs': {
                    'wf': {f.ELEMENT_KEY: MyFileUploader, 'type': 'wav',},
                    'func': {f.ELEMENT_KEY: sf.SelectBox, 'options': mall['func'],},
                },
            }
        },
    }
}


from streamlitfront import mk_app

app = mk_app([apply_func_to_wf], config=config_)
app()

# families = {'whalen': ['cora', 'thor', 'ness'], 'feron': ['valentin', 'maya', 'haydeé']}


# if 'mall' not in st.session_state:
#     st.session_state['mall'] = {
#         'family_name': dict(zip(families, families)),
#         'first_name': dict(),
#     }
#
# mall = st.session_state['mall']
# crudify = partial(prepare_for_crude_dispatch, mall=mall)

# locals = st.session_state['locals']


# stores = {
#     'fruit': {'apple': [1, 2, 3], 'banana': 'abc'},
#     'numbers': {'one': 1, 'two': 2},
# }

from dataclasses import dataclass
from typing import Any


# class Get:
#     key: str
#
#
# Nest = partial
#
#
# first_name_element = Nest(
#     st.selectbox,
#     options=Nest(families, Get('family_name'))
# )
#
# def get_value_from_store(family_name, first_name):
#     return f'{first_name} {family_name}'
#
#
# first_name_element.options = F(foo, family_name.value)


# ============ END BACKEND ============


# ============ FRONTEND ============


import os



import pytest
from pytest import fixture
from i2 import Sig

from plunk.sb.sig_compatibility.sig_comp import (
    transform_key,
    variadic_compatibility,
    variadics_from_sig,
)


def test_transform_key():
    d = {'key1': 12, 'key2': 23}
    func = lambda x: x.upper()
    result = transform_key(d, func)
    expected = {'KEY1': 12, 'KEY2': 23}
    assert result == expected


def test_variadics_from_sig():
    def f1(a, b, /, c, *args, k1=12):
        pass

    def f2(a, b, /, c, *args, k1=12, **kwargs):
        pass

    sig1 = Sig(f1)
    sig2 = Sig(f2)
    assert variadics_from_sig(sig1) == (True, False)
    assert variadics_from_sig(sig2) == (True, True)


def test_variadic_compatibility():
    vpk = True, False, True, True
    assert not variadic_compatibility(*vpk)
    vpk = True, True, True, True
    assert variadic_compatibility(*vpk) is None
    vpk = True, True, True, False
    assert not variadic_compatibility(*vpk)



# taken from https://github.com/aidanjungo/StreamlitDirectoryPicker/blob/main/directorypicker.py
import streamlit as st
from pathlib import Path


def st_directory_picker(initial_path=Path()):

    st.markdown('#### Directory picker')

    if 'path' not in st.session_state:
        st.session_state.path = initial_path.absolute()

    st.text_input('Selected directory:', st.session_state.path)

    _, col1, col2, col3, _ = st.columns([3, 1, 3, 1, 3])

    with col1:
        st.markdown('#')
        if st.button('⬅️') and 'path' in st.session_state:
            st.session_state.path = st.session_state.path.parent
            st.experimental_rerun()

    with col2:
        subdirectroies = [
            f.stem
            for f in st.session_state.path.iterdir()
            if f.is_dir()
            and (not f.stem.startswith('.') and not f.stem.startswith('__'))
        ]
        if subdirectroies:
            st.session_state.new_dir = st.selectbox(
                'Subdirectories', sorted(subdirectroies)
            )
        else:
            st.markdown('#')
            st.markdown(
                "<font color='#FF0000'>No subdir</font>", unsafe_allow_html=True
            )

    with col3:
        if subdirectroies:
            st.markdown('#')
            if st.button('➡️') and 'path' in st.session_state:
                st.session_state.path = Path(
                    st.session_state.path, st.session_state.new_dir
                )
                st.experimental_rerun()

    return st.session_state.path



import streamlit as st
from typing import List, Iterable
import soundfile as sf
from io import BytesIO
import numpy as np

Waveform = Iterable[int]


def tagged_sounds_to_single_array(train_audio: List[Waveform], tag: str):
    sounds, tag = train_audio, tag
    result = []
    for sound in sounds:
        # if not isinstance(sound, bytes):
        sound = sound.getvalue()
        arr = sf.read(BytesIO(sound), dtype='int16')[0]
        result.append(arr)
    print(np.hstack(result))
    return np.hstack(result), tag


st.title('Dummy app')
files = st.file_uploader(label='one', accept_multiple_files=True)
arr, tag = tagged_sounds_to_single_array(files, tag='Normal')
st.write(type(arr), arr.shape, arr[0])



"""
An app that loads either a wav file from local folder or records a sound
and visualizes the resulting numpy array 
"""
from typing import Mapping
from know.boxes import *
from functools import partial
from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY
from front.crude import Crudifier

from streamlitfront import mk_app, binder as b
from streamlitfront.elements import SuccessNotification
from streamlitfront.elements import FileUploader

from plunk.sb.front_demo.user_story1.components.components import ArrayPlotter
from plunk.sb.front_demo.user_story1.utils.tools import DFLT_PIPELINE
from plunk.sb.front_demo.user_story1.utils.simple_config import Component


def mk_pipeline_maker_app_with_mall(
    mall: Mapping,
    *,
    pipelines: str = None,
    sound_output: str = None,
    models_scores: str = None
):

    from plunk.sb.front_demo.user_story1.utils.funcs import (
        learn_apply_model,
        upload_sound,
    )

    if not b.mall():
        b.mall = mall
    mall = b.mall()

    crudifier = partial(Crudifier, mall=mall)

    learn_component = Component(func=learn_apply_model)
    learn_config = learn_component.mk_configs(
        {'execution.output': {ELEMENT_KEY: ArrayPlotter,},}
    )
    learn_apply_model = crudifier(
        param_to_mall_map=dict(tagged_data='sound_output', pipeline='pipelines',),
        output_store='models_scores',
    )(learn_apply_model)

    upload_component = Component(func=upload_sound)
    upload_config = upload_component.mk_configs(
        {
            'execution.inputs.train_audio': {
                ELEMENT_KEY: FileUploader,
                'type': 'wav',
                'accept_multiple_files': True,
            },
        }
    )

    upload_sound = crudifier(output_store='sound_output')(upload_sound)

    # do the whole config to view and compare with old-style one
    # config = {
    #     APP_KEY: {"title": "Data Preparation"},
    #     RENDERING_KEY: {
    #         "upload_sound": upload_config,
    #         "learn_apply_model": {
    #             "execution": {
    #                 "output": {
    #                     ELEMENT_KEY: ArrayPlotter,
    #                 },
    #             }
    #         },
    #     },
    # }

    config = {
        APP_KEY: {'title': 'Data Preparation'},
        RENDERING_KEY: {
            'upload_sound': {
                'execution': {
                    'inputs': {
                        'train_audio': {
                            ELEMENT_KEY: FileUploader,
                            'type': 'wav',
                            'accept_multiple_files': True,
                        },
                    },
                },
            },
            'learn_apply_model': {
                NAME_KEY: 'Apply model',
                'execution': {'output': {ELEMENT_KEY: ArrayPlotter,},},
            },
        },
    }

    funcs = [
        upload_sound,
        learn_apply_model,
    ]
    app = mk_app(funcs, config=config)

    return app


if __name__ == '__main__':

    mall = dict(sound_output=dict(), pipelines=dict(), models_scores=dict(),)

    mall['sound_output'] = dict()
    mall['pipelines'] = dict(
        # ML
        default_pipeline=DFLT_PIPELINE
    )

    app = mk_pipeline_maker_app_with_mall(
        mall, pipelines='pipelines', sound_output='sound_output'
    )

    app()



from hear import WavSerializationTrans, WavLocalFileStore
from forged.clickify import clickify, pad_to_length

# data prep
my_wav_store = WavLocalFileStore('./data/ClickDetection')
click_template = my_wav_store['base_click.wav']
wf_base = my_wav_store['wf_base.wav']
long_click = pad_to_length(click_template, len(wf_base))

# place clicks at arbitrary timestamps
timestamps = [2000, 10000, 20000, 40000, 60000]
wf_with_clicks = clickify(wf_base, timestamps, long_click, click_marker=None)


if __name__ == '__main__':

    print(len(wf_with_clicks))



"""Ideas for routing
See https://github.com/i2mint/i2/issues/40
"""

from typing import Mapping, Callable, TypeVar, KT, VT, Union
from i2 import mk_sentinel

Obj = TypeVar('Obj')
KeyFunc = Callable[[Obj], KT]
dflt_not_found_sentinel = mk_sentinel('dflt_not_found_sentinel')


def identity(x):
    return x


def key_func_mapping(
    obj: Obj,
    mapping: Mapping[KT, VT],
    key: KeyFunc = identity,
    not_found_sentinel=dflt_not_found_sentinel,
) -> VT:
    """Map an object to a value based on a key function"""
    return mapping.get(key(obj), not_found_sentinel)



from py2store import myconfigs
import numpy as np
from functools import partial
from sklearn.preprocessing import normalize
from hear import WavLocalFileStore
from dol import wrap_kvs
import soundfile as sf
from io import BytesIO
from odat.utils.chunkers import fixed_step_chunker

DFLT_CHUNKER = partial(fixed_step_chunker, chk_size=2048)

config_filename = 'guns.json'
DFLT_LOCAL_SOURCE_DIR = myconfigs.get_config_value(config_filename, 'local_source_dir')


def mk_dacc(root_dir=DFLT_LOCAL_SOURCE_DIR):
    return Dacc(root_dir=root_dir)


def wf_from_bytes(bytes):
    return sf.read(BytesIO(bytes), dtype='float32')[0]


def WfStore(root_store):
    obj_of_data = wf_from_bytes
    return wrap_kvs(root_store, obj_of_data=obj_of_data)


class Dacc:
    def __init__(self, root_dir=DFLT_LOCAL_SOURCE_DIR):
        self.wfs = WavLocalFileStore(root_dir)

    def wf_tag_gen(self):
        for key in self.wfs:
            signal = self.wfs[key]
            normal_wf = normalize(np.float32(signal).reshape(1, -1))[0]

            yield normal_wf, key

    def chk_tag_gen(self, chunker=DFLT_CHUNKER):
        for wf, tag in self.wf_tag_gen():
            for chk in chunker(wf):
                yield chk, tag


if __name__ == '__main__':
    dacc = mk_dacc()
    print(next(dacc.chk_tag_gen()))



import websocket
import json
from streamz import Stream
from streamz.dataframe import DataFrame
import pandas as pd
import hvplot.streamz
from streamz.dataframe import Random
import datetime
import plotly.graph_objects as go

source = Stream()

df = pd.DataFrame({'x': []})
sdf = DataFrame(source, example=df)


def on_message(ws, message):
    values = json.loads(message)['values']
    x = values[0]
    y = values[1]
    z = values[2]
    source.emit(pd.DataFrame({'x': [float(x)]}, index=[pd.Timestamp.now()]))
    # source.emit(pd.DataFrame({"x": [x]},index=[datetime.datetime.now()]))


#     pd.DataFrame(dict(used=100*vmem.used/vmem.total,
#                              free=100*vmem.free/vmem.total),
#                         index=[pd.Timestamp.now()])
#     t = datetime.datetime.now()
#     times.append(t)
#     vals.append(float(x))
#     fig.data[0].x = times
#     fig.data[0].y = vals
# df = df.append(pd.DataFrame({"x": float(x)}, index=[t]))
# print(sdf.x.sum())


def on_error(ws, error):
    print('error occurred')
    print(error)


def on_close(ws, close_code, reason):
    print('connection close')
    print('close code : ', close_code)
    print('reason : ', reason)


def on_open(ws):
    print('connection open')


if __name__ == '__main__':
    ws = websocket.WebSocketApp(
        'ws://192.168.1.3:8080/sensor/connect?type=android.sensor.accelerometer',
        on_open=on_open,
        on_message=on_message,
        on_error=on_error,
        on_close=on_close,
    )
    sdf.hvplot(backlog=100)
    # sdf.hvplot()

    # fig.show()

    ws.run_forever()

# To analyse multiple sensor data simultaneously, you can add as many websocket connections for different sensors as you want.



"""An app that loads wav file from local folder"""
import re
from functools import partial, reduce, wraps
from typing import Iterable

from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY

from front.crude import Crudifier

from streamlitfront import mk_app, binder as b
from streamlitfront.elements import SuccessNotification, SelectBox
from streamlitfront.elements import FileUploader

from plunk.ap.store_explorer.store_explorer_element import (
    StoreExplorerOutput,
    StoreExplorerInput,
)
from plunk.ap.snippets import get_mall


def tuple_wrap(f):
    @wraps(f)
    def _wrap(*a, **kw):
        return (), f(*a, **kw)

    return _wrap


def mk_pipeline_maker_app_with_mall(mall: dict):
    mall = get_mall(mall)

    def on_select_mall_name():
        print(f'{b.selected_mall_name()=}')
        b.selected_mall.set(mall[b.selected_mall_name()])

    if not b.mall_names():
        b.mall_names = lambda: list(mall)
        b.selected_mall_name.set(next(iter(b.mall_names()())))
        on_select_mall_name()
    crudifier = partial(Crudifier, mall=mall)

    @crudifier(output_store='sound_output')
    def upload_sound(train_audio: list, tag: str):
        return train_audio, tag

    def explore_mall(mall_name, depth_keys: Iterable = ()):
        return depth_keys, reduce(lambda o, k: o[k], depth_keys, mall[mall_name])

    @tuple_wrap
    def search_mall(
        key_filter: str,
        match_case: bool = False,
        match_whole_key: bool = False,
        use_regex: bool = False,
    ):
        """By vf"""

        def filter_dict_in_depth(d, predicate):
            def build_filtered_dict():
                for k, v in d.items():
                    if predicate(k):
                        yield k, v
                    elif isinstance(v, dict):
                        _v = filter_dict_in_depth(v, predicate)
                        if _v:
                            yield k, _v

            return {k: v for k, v in build_filtered_dict()}

        if not key_filter:
            return mall
        if not match_case:
            key_filter = key_filter.lower()
        if use_regex:

            def include_key(key):
                if not match_case:
                    key = key.lower()
                regex = f'^{key_filter}$' if match_whole_key else key_filter
                if re.search(regex, str(key)):
                    return True
                return False

        else:

            def include_key(key):
                if not match_case:
                    key = key.lower()
                if match_whole_key:
                    return key_filter == key
                return key_filter in str(key)

        return filter_dict_in_depth(mall, include_key)

    config = {
        APP_KEY: {'title': 'Data Preparation'},
        RENDERING_KEY: {
            'upload_sound': {
                # NAME_KEY: "Data Loader",
                'execution': {
                    'inputs': {
                        'train_audio': {
                            ELEMENT_KEY: FileUploader,
                            'type': 'wav',
                            'accept_multiple_files': True,
                        },
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'Wav files loaded successfully.',
                    },
                },
            },
            'explore_mall': {
                'execution': {
                    'inputs': {
                        'mall_name': {
                            ELEMENT_KEY: SelectBox,
                            'options': b.mall_names(),
                            'value': b.selected_mall_name,
                            'on_value_change': on_select_mall_name,
                        },
                        'depth_keys': {
                            ELEMENT_KEY: StoreExplorerInput,
                            'mall': b.selected_mall,
                        },
                    },
                    'output': {
                        ELEMENT_KEY: StoreExplorerOutput,
                        'write_depth_keys': True,
                    },
                    'auto_submit': True,
                },
            },
            'search_mall': {
                'execution': {
                    'output': {
                        ELEMENT_KEY: StoreExplorerOutput,
                        'write_depth_keys': False,
                    },
                    'auto_submit': True,
                },
            },
        },
    }

    funcs = [
        explore_mall,
        search_mall,
        upload_sound,
    ]
    app = mk_app(funcs, config=config)

    return app


_mall = dict(
    # Factory Input Stores
    sound_output=dict(),
    # Output Store
    data=dict(
        a=1,
        b='2',
        c=[3],
        d={'4': 4, '5': [{'e': 6}, {'f': 7}, ['g', 'h', 'i']]},
        j=[{'k': 8, 'l': 9}, 'm', 10],
    ),
)


if __name__ == '__main__':

    app = mk_pipeline_maker_app_with_mall(_mall)

    app()
